#ifndef __DRAWATOM_H
#define __DRAWATOM_H


/*
************************************************************************
File Includes
************************************************************************
*/
#include <cor0.h>


/*
************************************************************************
Defines
************************************************************************
*/

#define BOOLEAN int
#define FALSE 0
#define TRUE 1

#define WORD2 unsigned short

/*
Maximum number of colors assignments
*/
#define MAX_COLOR_MAP 256

/*
Maximum number of radii assignments
*/
#define MAX_RADIUS_MAP 256


/*  Directions  */
#define NDIR 3
#define X 0
#define Y 1
#define Z 2


/*  Special color value denoting clear color (i.e. drawing done)	*/
#define CLEAR_COLOR 0xffffU

/*  Define far only for PC machines  */
#ifdef __TURBOC__
#define FAR far
#else
#define FAR
#endif




/*
************************************************************************
Type Definitions
************************************************************************
*/

typedef struct
	{
	BOOLEAN AtomFlag;
	BOOLEAN AtomIndexFlag;
	BOOLEAN BondFlag;
	BOOLEAN BondWrapFlag;
	BOOLEAN ColorAtomByDispAngleFlag;
	BOOLEAN ColorAtomByDispMagFlag;
	BOOLEAN ColorAtomByInplaneDispMagFlag;
	BOOLEAN ColorAtomByDispVectFlag;
	BOOLEAN ColorAtomByRelDispMagFlag;
	BOOLEAN ColorAtomByInplaneRelDispMagFlag;
	BOOLEAN ColorAtomByRelDispVectFlag;
	BOOLEAN ColorAtomByTypeFlag;
	BOOLEAN ColorAtomByLayerFlag;
	BOOLEAN ColorBondByDistFlag;
	BOOLEAN DispFlag;
	BOOLEAN DispRefFlag;
	BOOLEAN LabelFlag;
	BOOLEAN LayerRefFlag;
	BOOLEAN RadiusAtomByTypeFlag;
	BOOLEAN ResolutionFlag;
	BOOLEAN TailFlag;

	char	 *TypeListFileName;
	char	 *DispRefFileName;
	char	 *LayerRefFileName;

	char	 **Title;
	int	 NumTitle;

	BOOLEAN TailSlipFlag;
	float   TailSlipRad;
	float   TailScale;
	BOOLEAN TypeListFlag;

	int hor, ver, dep; /*  COORDINATES DIRECTIONS IN SCREEN FRAME OF REF  */
	int ilayer, nlayer;/*  LAYER INFO												 */
	float cmin[NDIR], cmax[NDIR];         /*	PARTICLE COORDINATE LIMITS  */
	int	hsmin,  hsmax,  vsmin,	vsmax;  /*	SCREEN	COORDINATE LIMITS  */
	int	tmarg,  bmarg,  lmarg,	rmarg;  /*	MARGINS (IN PIXELS)			 */
					 /*  NOT USED IF labelflag FALSE */
	int	timarg, bimarg, limarg, rimarg; /*	INNVER MARGINS (IN PIXELS)  */
					 /*  SPACE BETWEEN BORDER AND PIC*/
	float bnd1, bnd2; 						  /*	BOND LIMITS 					 */
	float aspect;								  /*	SCREEN ASPECT RATIO			 */

	int	vectflag;
	int	vectrelflag;
	/* Vector for reference for colors */
	float vector[NDIR];
	float vectmin, vectmax;
	float DispMag;
	float Phase;

	/*  Number of displacements colors	*/
	int	NumDispColor;

	/*  Number of bond colors	*/
	int	NumBondColor;

	/*  Number of layer colors  */
	int	NumLayerColor;

	/*  Atom Colors for different atom types	*/
	WORD2 AtomColorMap [MAX_COLOR_MAP];

	/*  Bond Colors for different bonds  */
	WORD2 BondColorMap [MAX_COLOR_MAP];

	/*  Atom Radii for different atom types	*/
	double AtomRadiusMap [MAX_RADIUS_MAP];

	/*  Rotational matrix for view  */
	float Rotate[NDIR][NDIR];

	/*  Flag for arbitrary viewpoint direction  */
	BOOLEAN ViewFlag;

	/*  Frame for scaling view  */
	float Frame[2][NDIR];

	/*  Flag for turning frame on  */
	BOOLEAN FixedFrameFlag;

	} dsparm;


/*
************************************************************************
Function Prototypes
************************************************************************
*/

int  DS_DrawStep	(stepdata *, stepdata *, dsparm *);
void DS_DrawLabel (stepdata *, dsparm *);
void DS_SetFrameColor (WORD2);
void DS_SetTextColor  (WORD2);
void DS_SetDispColor  (WORD2);
void DS_SetBondColor  (WORD2);

#endif
