/*
SVGA Graphics Routines for 256 color mode
  Variable resolution
*/


/*
************************************************************************
File Includes
************************************************************************
*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>
#include "hlgraph.h"
#include "llgraph.h"

#ifdef __TURBOC__
#include <conio.h>
#endif


/*
************************************************************************
Defines
************************************************************************
*/
#define ULONG unsigned long
#define BYTE  unsigned char
#define MAXLEN 1024
#define DEFAULT_BACKGROUND_COLOR 15
#define MAX_PAGES 2



/*
***********************************************************************
Macros
***********************************************************************
*/


#define ALLOCATE(VAR,TYPE,NUM) \
   VAR = (TYPE *) calloc ( (NUM), sizeof ( TYPE ) ); \
   if (VAR == NULL) \
      { \
      GR_SetTextMode(); \
      printf ("ERROR (File %s Line %i):\n", __FILE__, __LINE__); \
      printf ("   Cannot allocate variable %s.\n", #VAR); \
      exit (1); \
      }

#define FREE(VAR) \
   if ((VAR)!=NULL) \
      { \
      free (VAR); \
      (VAR) = NULL; \
      }



/*
************************************************************************
External (Global) Variables
************************************************************************
*/
extern WORD2 HorizontalResolution_g;
extern WORD2 VerticalResolution_g;
extern WORD2 BackgroundColor_g;


/*
************************************************************************
Module-wide variables
************************************************************************
*/


/*
Character bit masks for ascii 0..127
*/
static char CharTable_m[128][8] =
{
   { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
   { 0x7C, 0x82, 0xAA, 0x82, 0xBA, 0x92, 0x82, 0x7C },
   { 0x7C, 0xFE, 0xD6, 0xFE, 0xC6, 0xEE, 0xFE, 0x7C },
   { 0x6C, 0xEE, 0xFE, 0xFE, 0xFE, 0x7C, 0x38, 0x10 },
   { 0x10, 0x38, 0x7C, 0xFE, 0xFE, 0x7C, 0x38, 0x10 },
   { 0x38, 0x38, 0x38, 0xFE, 0xFE, 0xEE, 0x10, 0x38 },
   { 0x10, 0x38, 0x7C, 0xFE, 0xFE, 0x66, 0x10, 0x38 },
   { 0x00, 0x00, 0x38, 0x7C, 0x7C, 0x38, 0x00, 0x00 },
   { 0xFF, 0xFF, 0xC7, 0x83, 0x83, 0xC7, 0xFF, 0xFF },
   { 0x00, 0x3C, 0x66, 0x42, 0x42, 0x66, 0x3C, 0x00 },
   { 0xFF, 0xC3, 0x99, 0xBD, 0xBD, 0x99, 0xC3, 0xFF },
   { 0x3E, 0x0E, 0x1A, 0x32, 0x78, 0xCC, 0xCC, 0x78 },
   { 0x3C, 0x66, 0x66, 0x3C, 0x18, 0x7E, 0x18, 0x18 },
   { 0x1C, 0x1E, 0x1A, 0x18, 0x38, 0x78, 0x70, 0x00 },
   { 0x78, 0x6C, 0x66, 0x66, 0xE6, 0xE6, 0x0E, 0x0E },
   { 0x10, 0x54, 0x38, 0xEE, 0x38, 0x54, 0x10, 0x00 },
   { 0x60, 0x70, 0x78, 0x7C, 0x78, 0x70, 0x60, 0x00 },
   { 0x0C, 0x1C, 0x3C, 0x7C, 0x3C, 0x1C, 0x0C, 0x00 },
   { 0x10, 0x38, 0x7C, 0x10, 0x7C, 0x38, 0x10, 0x00 },
   { 0xC6, 0xC6, 0xC6, 0xC6, 0x00, 0xC6, 0xC6, 0x00 },
   { 0x7E, 0xD6, 0xD6, 0xD6, 0x76, 0x16, 0x16, 0x00 },
   { 0x3C, 0x66, 0x78, 0x24, 0x1E, 0x66, 0x3C, 0x00 },
   { 0x00, 0x00, 0x00, 0xFE, 0xFE, 0xFE, 0xFE, 0x00 },
   { 0x10, 0x38, 0x7C, 0x10, 0x7C, 0x38, 0x10, 0x7C },
   { 0x18, 0x3C, 0x7E, 0x18, 0x18, 0x18, 0x18, 0x00 },
   { 0x18, 0x18, 0x18, 0x18, 0x7E, 0x3C, 0x18, 0x00 },
   { 0x00, 0x18, 0x0C, 0xFE, 0x0C, 0x18, 0x00, 0x00 },
   { 0x00, 0x30, 0x60, 0xFE, 0x60, 0x30, 0x00, 0x00 },
   { 0x00, 0xC0, 0xC0, 0xC0, 0xC0, 0xFE, 0x00, 0x00 },
   { 0x00, 0x28, 0x6C, 0xFE, 0x6C, 0x28, 0x00, 0x00 },
   { 0x00, 0x10, 0x38, 0x7C, 0xFE, 0x00, 0x00, 0x00 },
   { 0x00, 0xFE, 0x7C, 0x38, 0x10, 0x00, 0x00, 0x00 },
   { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
   { 0x30, 0x30, 0x30, 0x30, 0x00, 0x30, 0x30, 0x00 },
   { 0xC6, 0xC6, 0xC6, 0x00, 0x00, 0x00, 0x00, 0x00 },
   { 0x6C, 0x6C, 0xFE, 0x6C, 0xFE, 0x6C, 0x6C, 0x00 },
   { 0x18, 0x7E, 0xC0, 0x7C, 0x06, 0xFC, 0x18, 0x00 },
   { 0xC4, 0xCC, 0x18, 0x30, 0x60, 0xCC, 0x8C, 0x00 },
   { 0x38, 0x6C, 0x38, 0x76, 0xDC, 0xCC, 0x76, 0x00 },
   { 0x18, 0x18, 0x30, 0x00, 0x00, 0x00, 0x00, 0x00 },
   { 0x0C, 0x18, 0x30, 0x30, 0x30, 0x18, 0x0C, 0x00 },
   { 0x30, 0x18, 0x0C, 0x0C, 0x0C, 0x18, 0x30, 0x00 },
   { 0x00, 0x6C, 0x38, 0xFE, 0x38, 0x6C, 0x00, 0x00 },
   { 0x00, 0x30, 0x30, 0xFC, 0x30, 0x30, 0x00, 0x00 },
   { 0x00, 0x00, 0x00, 0x00, 0x00, 0x18, 0x18, 0x30 },
   { 0x00, 0x00, 0x00, 0xFC, 0x00, 0x00, 0x00, 0x00 },
   { 0x00, 0x00, 0x00, 0x00, 0x00, 0x18, 0x18, 0x00 },
   { 0x02, 0x06, 0x0C, 0x18, 0x30, 0x60, 0x40, 0x00 },
   { 0x7C, 0xC6, 0xCE, 0xDE, 0xF6, 0xE6, 0x7C, 0x00 },
   { 0x30, 0x70, 0xF0, 0x30, 0x30, 0x30, 0xFC, 0x00 },
   { 0x78, 0xCC, 0xCC, 0x18, 0x30, 0x60, 0xFC, 0x00 },
   { 0xFC, 0x0C, 0x18, 0x38, 0x0C, 0xCC, 0x78, 0x00 },
   { 0x1C, 0x3C, 0x6C, 0xCC, 0xFE, 0x0C, 0x0C, 0x00 },
   { 0xFC, 0xC0, 0xC0, 0xF8, 0x0C, 0xCC, 0x78, 0x00 },
   { 0x3C, 0x60, 0xC0, 0xFC, 0xC6, 0xC6, 0x7C, 0x00 },
   { 0xFE, 0xC6, 0x0C, 0x18, 0x30, 0x30, 0x30, 0x00 },
   { 0x7C, 0xC6, 0xC6, 0x7C, 0xC6, 0xC6, 0x7C, 0x00 },
   { 0x7C, 0xC6, 0xC6, 0x7E, 0x06, 0x0C, 0x78, 0x00 },
   { 0x00, 0x18, 0x18, 0x00, 0x00, 0x18, 0x18, 0x00 },
   { 0x00, 0x18, 0x18, 0x00, 0x00, 0x18, 0x18, 0x30 },
   { 0x0C, 0x18, 0x30, 0x60, 0x30, 0x18, 0x0C, 0x00 },
   { 0x00, 0x00, 0xFC, 0x00, 0xFC, 0x00, 0x00, 0x00 },
   { 0x60, 0x30, 0x18, 0x0C, 0x18, 0x30, 0x60, 0x00 },
   { 0x78, 0xCC, 0xCC, 0x18, 0x30, 0x00, 0x30, 0x00 },
   { 0x7C, 0xC6, 0xCE, 0xD6, 0xDE, 0xC0, 0x78, 0x00 },
   { 0x38, 0x6C, 0xC6, 0xC6, 0xFE, 0xC6, 0xC6, 0x00 },
   { 0xFC, 0x66, 0x66, 0x7C, 0x66, 0x66, 0xFC, 0x00 },
   { 0x7C, 0xC6, 0xC0, 0xC0, 0xC0, 0xC6, 0x7C, 0x00 },
   { 0xFC, 0x66, 0x66, 0x66, 0x66, 0x66, 0xFC, 0x00 },
   { 0xFC, 0xC0, 0xC0, 0xF8, 0xC0, 0xC0, 0xFC, 0x00 },
   { 0xFC, 0xC0, 0xC0, 0xF8, 0xC0, 0xC0, 0xC0, 0x00 },
   { 0x7C, 0xC6, 0xC0, 0xDE, 0xC6, 0xC6, 0x7E, 0x00 },
   { 0xC6, 0xC6, 0xC6, 0xFE, 0xC6, 0xC6, 0xC6, 0x00 },
   { 0xFC, 0x30, 0x30, 0x30, 0x30, 0x30, 0xFC, 0x00 },
   { 0x3E, 0x0C, 0x0C, 0x0C, 0x0C, 0xCC, 0x78, 0x00 },
   { 0xC6, 0xCC, 0xD8, 0xF0, 0xD8, 0xCC, 0xC6, 0x00 },
   { 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0xFC, 0x00 },
   { 0xC6, 0xEE, 0xFE, 0xD6, 0xC6, 0xC6, 0xC6, 0x00 },
   { 0xC6, 0xE6, 0xF6, 0xDE, 0xCE, 0xC6, 0xC6, 0x00 },
   { 0x7C, 0xC6, 0xC6, 0xC6, 0xC6, 0xC6, 0x7C, 0x00 },
   { 0xFC, 0xC6, 0xC6, 0xFC, 0xC0, 0xC0, 0xC0, 0x00 },
   { 0x7C, 0xC6, 0xC6, 0xC6, 0xF6, 0xDE, 0x7C, 0x06 },
   { 0xFC, 0xC6, 0xC6, 0xFC, 0xD8, 0xCC, 0xC6, 0x00 },
   { 0x7C, 0xC6, 0xC0, 0x7C, 0x06, 0xC6, 0x7C, 0x00 },
   { 0xFC, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x00 },
   { 0xC6, 0xC6, 0xC6, 0xC6, 0xC6, 0xC6, 0x7C, 0x00 },
   { 0xC6, 0xC6, 0xC6, 0xC6, 0x6C, 0x38, 0x10, 0x00 },
   { 0xC6, 0xC6, 0xC6, 0xD6, 0xFE, 0xEE, 0x44, 0x00 },
   { 0xC6, 0xC6, 0x6C, 0x38, 0x6C, 0xC6, 0xC6, 0x00 },
   { 0xCC, 0xCC, 0xCC, 0x78, 0x30, 0x30, 0x30, 0x00 },
   { 0xFC, 0x0C, 0x18, 0x30, 0x60, 0xC0, 0xFC, 0x00 },
   { 0x7C, 0x60, 0x60, 0x60, 0x60, 0x60, 0x7C, 0x00 },
   { 0x80, 0xC0, 0x60, 0x30, 0x18, 0x0C, 0x04, 0x00 },
   { 0x7C, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x7C, 0x00 },
   { 0x10, 0x38, 0x6C, 0xC6, 0x00, 0x00, 0x00, 0x00 },
   { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFE },
   { 0x18, 0x30, 0x30, 0x00, 0x00, 0x00, 0x00, 0x00 },
   { 0x00, 0x00, 0x78, 0x0C, 0x7C, 0xCC, 0x7C, 0x00 },
   { 0xC0, 0xC0, 0xD8, 0xEC, 0xCC, 0xCC, 0xF8, 0x00 },
   { 0x00, 0x00, 0x7C, 0xC0, 0xC0, 0xC0, 0x7C, 0x00 },
   { 0x0C, 0x0C, 0x6C, 0xDC, 0xCC, 0xCC, 0x7C, 0x00 },
   { 0x00, 0x00, 0x78, 0xCC, 0xFC, 0xC0, 0x7C, 0x00 },
   { 0x1C, 0x30, 0x30, 0x7C, 0x30, 0x30, 0x30, 0x00 },
   { 0x00, 0x00, 0x7C, 0xCC, 0xCC, 0x7C, 0x0C, 0x78 },
   { 0xC0, 0xC0, 0xD8, 0xEC, 0xCC, 0xCC, 0xCC, 0x00 },
   { 0x30, 0x00, 0x70, 0x30, 0x30, 0x30, 0x78, 0x00 },
   { 0x0C, 0x00, 0x1C, 0x0C, 0x0C, 0x0C, 0x6C, 0x38 },
   { 0xC0, 0xC0, 0xCC, 0xD8, 0xF0, 0xD8, 0xCC, 0x00 },
   { 0x70, 0x30, 0x30, 0x30, 0x30, 0x30, 0x78, 0x00 },
   { 0x00, 0x00, 0xCC, 0xFE, 0xFE, 0xD6, 0xC6, 0x00 },
   { 0x00, 0x00, 0xD8, 0xEC, 0xCC, 0xCC, 0xCC, 0x00 },
   { 0x00, 0x00, 0x78, 0xCC, 0xCC, 0xCC, 0x78, 0x00 },
   { 0x00, 0x00, 0xF8, 0xCC, 0xCC, 0xF8, 0xC0, 0xC0 },
   { 0x00, 0x00, 0x7C, 0xCC, 0xCC, 0x7C, 0x0C, 0x0E },
   { 0x00, 0x00, 0xD8, 0xEC, 0xC0, 0xC0, 0xC0, 0x00 },
   { 0x00, 0x00, 0x7C, 0xC0, 0x78, 0x0C, 0xF8, 0x00 },
   { 0x00, 0x60, 0xF8, 0x60, 0x60, 0x6C, 0x38, 0x00 },
   { 0x00, 0x00, 0xCC, 0xCC, 0xCC, 0xDC, 0x6C, 0x00 },
   { 0x00, 0x00, 0xC6, 0x6C, 0x6C, 0x38, 0x10, 0x00 },
   { 0x00, 0x00, 0xC6, 0xD6, 0xFE, 0xFE, 0x6C, 0x00 },
   { 0x00, 0x00, 0xCC, 0x78, 0x30, 0x78, 0xCC, 0x00 },
   { 0x00, 0x00, 0xCC, 0xCC, 0xCC, 0x7C, 0x0C, 0x78 },
   { 0x00, 0x00, 0xFC, 0x18, 0x30, 0x60, 0xFC, 0x00 },
   { 0x1C, 0x30, 0x30, 0x60, 0x30, 0x30, 0x1C, 0x00 },
   { 0x18, 0x18, 0x18, 0x00, 0x18, 0x18, 0x18, 0x00 },
   { 0x70, 0x18, 0x18, 0x0C, 0x18, 0x18, 0x70, 0x00 },
   { 0x66, 0xD6, 0xCC, 0x00, 0x00, 0x00, 0x00, 0x00 },
   { 0x00, 0x00, 0x10, 0x38, 0x6C, 0xC6, 0xFE, 0x00 }
};



/*
************************************************************************
Exported routines
************************************************************************
*/


void GR_Ver (WORD2 x, WORD2 y1, WORD2 y2, WORD2 Color)
   {
   WORD2 ylimit;
   WORD2 y;


   /*  Order y coordinates   */
   if (y1<y2)
      {
      y      = y1;
      ylimit = y2;
      }
   else
      {
      y      = y2;
      ylimit = y1;
      }

   /*  Write pixels  */
   while (y<=ylimit)
      {
      GR_WritePixel (x, y, Color);
      y++;
      }
   }


void GR_Hor (WORD2 y, WORD2 x1, WORD2 x2, WORD2 Color)
   {
   WORD2 xlimit;
   WORD2 x;

   /*  Order x coordinates   */
   if (x1<x2)
      {
      x      = x1;
      xlimit = x2;
      }
   else
      {
      x      = x2;
      xlimit = x1;
      }

   /*  Write piyels  */
   while (x<=xlimit)
      {
      GR_WritePixel (x, y, Color);
      x++;
      }
   }


void GR_Box   (WORD2 x1, WORD2 y1, WORD2 x2, WORD2 y2, WORD2 Color)
   {
   GR_Ver (x1, y1, y2, Color);
   GR_Ver (x2, y1, y2, Color);
   GR_Hor (y1, x1, x2, Color);
   GR_Hor (y2, x1, x2, Color);
   }

void GR_Atom (WORD2 InputX, WORD2 InputY, WORD2 Color)
   {
   int i;
   int Count;
   WORD2 x;
   WORD2 y;

   char Image[49] =
      {
      0,0,0,0,0,0,0,
      0,0,1,1,1,0,0,
      0,1,1,1,1,1,0,
      0,1,1,1,1,1,0,
      0,1,1,1,1,1,0,
      0,0,1,1,1,0,0,
      0,0,0,0,0,0,0
      };
   char Mask[49] =
      {
      0,0,1,1,1,0,0,
      0,1,0,0,0,1,0,
      1,0,0,0,0,0,1,
      1,0,0,0,0,0,1,
      1,0,0,0,0,0,1,
      0,1,0,0,0,1,0,
      0,0,1,1,1,0,0
      };

   /*  Initialize coordinates  */
   x = InputX-3;
   y = InputY-3;

   /*  Write mask  */
   Count  = 0;
   for (i=0; i<49; i++)
      {
      if (Mask[i])
         GR_WritePixel (x, y, BackgroundColor_g);
      Count++;
      x++;
      if (Count==7)
         {
         Count=0;
         x -= 7;
         y ++;
         }
      }

   /*  Initialize coordinates  */
   x = InputX-3;
   y = InputY-3;


   /*  Write mask  */
   Count  = 0;
   for (i=0; i<49; i++)
      {
      if (Image[i])
         GR_WritePixel (x, y, Color);
      Count++;
      x++;
      if (Count==7)
         {
         Count=0;
         x -= 7;
         y ++;
         }
      }
   }


void GR_Line  (WORD2 x1, WORD2 y1, WORD2 x2, WORD2 y2, WORD2 Color)
   {
   int ydel;
   int xdel;
   int n;
   int inc;
   int cum;
   WORD2 x;
   WORD2 y;

   /*  ORDER BY X  */
   if (x1>x2)
      {
      n  = x2;
      x2 = x1;
      x1 = n;
      n  = y2;
      y2 = y1;
      y1 = n;
      }

   /*  GET Y INCREMENT  */
   if (y1>y2)
      inc = -1;
   else
      inc =  1;

   /*  X AND Y SPANS  */
   xdel =      x2-x1;     /*  x's ordered so fabs unecessary  */
   ydel =      y2-y1;
   if (ydel<0)
      ydel = -ydel;

   /*  Set starting point  */
   x = x1;
   y = y1;
   GR_WritePixel(x, y, Color);

   /*  STEP IN +x DIRECTION  */
   if (xdel>ydel)
      {
      cum = xdel >> 1;
      n = x2-x1;
      while (n--)
         {
         /*  Increment x  */
         x++;
         cum += ydel;
         if (cum>xdel)
            {
            cum -= xdel;
            y += inc;
            }
         GR_WritePixel(x, y, Color);
         }
      }

   /*  STEP IN <inc> y DIRECTION  */
   else
      {
      cum = ydel >> 1;
      if (y2>y1)
         n = y2-y1;
      else
         n = y1-y2;
      while (n--)
         {
         y += inc;
         cum += xdel;
         if (cum>ydel)
            {
            cum -= ydel;
            x++;
            }
         GR_WritePixel(x, y, Color);
         }
      }
   }



void GR_Char  (WORD2 x, WORD2 y, WORD2 Color, char c)
   {
   char *CharPtr =
      (char *) CharTable_m + ( ( (BYTE) c & 0x007f) << 3);
   char OutputMask;
   int NumRows;
   int NumCols;
   WORD2 ForegroundColor = Color;

   NumRows = 8;
   while (NumRows--)
      {

      /*  Get current line of character  */
      OutputMask = *(CharPtr++);

      /*  Scan of 8 columns  */
      NumCols = 8;
      while (NumCols--)
         {

         /*  Bit is on, write color  */
         if (OutputMask & 0x80)
            {
            GR_WritePixel(x, y, ForegroundColor);
            }

         /*  Bit is off, write background  */
         else
            {
            GR_WritePixel(x, y, BackgroundColor_g);
            }

         /*  Increment pixel position  */
         x++;

         /*  Shift character mask  */
         OutputMask <<= 1;
         }


      /*  Move pixel to next row  */
      x -= 8;
      y ++;

      }

   }


/*
Print string on graphics screen
 - wraps text in x direction
 - if x>LimitX move to beginning of next line
 - if y>LmiitY overwrite last line
*/
void GR_Printf (WORD2 x, WORD2 y, WORD2 Color, char *format, ...)
   {
   char cbuffer[MAXLEN];
   va_list argptr;
   char *cp = cbuffer;
   WORD2 LimitX;
   WORD2 LimitY;

   /*  Initialize limits  */
   LimitX = HorizontalResolution_g - 8;
   LimitY = VerticalResolution_g   - 8;

   /*  CONVERT ARGUMENTS TO STRING  */
   va_start (argptr, format);
   vsprintf (cbuffer, format, argptr);
   va_end   (argptr);

   /*  WRITE OUT STRING  */
   while (*cp!=0)
      {

      /*  If x is beyond right edge, continue on next line  */
      if (x>=LimitX)
         {
         x  = 0;
         y += 8;
         }

      /*  If y position below screen, write it on bottom line */
      if (y>LimitY)
         y = LimitY;

      /*  Call character routine  */
      GR_Char (x, y, Color, *(cp++));

      /*  Increment x position for next character  */
      x += 8;
      }
   }


void GR_Compress
(
BYTE *InputBitMap,
BYTE *CompressedBitMap,
WORD2 InputSize,
WORD2 *CompressedSize
)
   {
   BYTE *InputPtr;
   WORD2 TotalRepeatFactor;
   WORD2 RemainingRepeatFactor;
   WORD2 PartialRepeatFactor;
   WORD2 RemainingMoveFactor;
   WORD2 CurMoveFactor;
   WORD2 NumCompressedBytes;
   WORD2 MaxCompressedSize;
   BYTE CurByte;
   BYTE NextByte;
   BYTE *InputPtrLimit = (BYTE *) &(InputBitMap[ InputSize]);
   BYTE *CompressedBitMapLimit = (BYTE *) &(CompressedBitMap[*CompressedSize]);

   MaxCompressedSize = *CompressedSize;
   NumCompressedBytes = 0;
   do
      {

      /*  Determine number of repeating bytes  */
      InputPtr = InputBitMap;
      CurByte  = *(InputPtr++);
      TotalRepeatFactor = 1;
      while (InputPtr<InputPtrLimit && *(InputPtr++)==CurByte)
         TotalRepeatFactor++;

      /*  Code repeating bytes in groups of 128 or less  */
      RemainingRepeatFactor = TotalRepeatFactor;
      while (RemainingRepeatFactor>2)
         {

         /*  Limit number of coded repeating bytes to 128  */
         if (RemainingRepeatFactor>=128)
           PartialRepeatFactor = 128;
         else
           PartialRepeatFactor = RemainingRepeatFactor;

         /*  Determine number of left over repeating bytes  */
         RemainingRepeatFactor -= PartialRepeatFactor;

         /*  Tally number of bytes of output code  */
         NumCompressedBytes += 2;

         /*  Is total of output bytes too big?  */
         if (NumCompressedBytes>MaxCompressedSize)
            {
            fprintf (stderr, "ERROR (Gcompress): Not enough room for new map.\n");
            exit(1);
            }

         /*  Write code to output map  */
         *(CompressedBitMap++) = (PartialRepeatFactor-1) | 0x80;
         *(CompressedBitMap++) = CurByte;

         }  /*  Remaining Repeat Factor > 2  */

      /*  Point to input past repeating bytes   */
      InputBitMap    += TotalRepeatFactor-RemainingRepeatFactor;

      /*
      Search for bytes to move
      Move subsequent bytes if three or more unique consecutive bytes
      */

      /*  Temporary pointer to input buffer  */
      InputPtr          = InputBitMap;

      /*  Search for string of unlike characters  */
      CurByte  = *(InputPtr++);
      NextByte = *(InputPtr++);
      RemainingMoveFactor = 0;
      while
      (
      InputPtr<InputPtrLimit &&
      ( CurByte!=NextByte || NextByte!=(*InputPtr))
      )
         {
         CurByte = NextByte;
         NextByte = *(InputPtr++);
         RemainingMoveFactor++;
         }

      /*  If at end of input buffer, include last two bytes  */
      if (InputPtr >= InputPtrLimit)
         RemainingMoveFactor += 2;

      /*  Write coded bytes in size block of 128 or less  */
      while (RemainingMoveFactor>0)
         {
         if (RemainingMoveFactor>=128)
           CurMoveFactor = 128;
         else
           CurMoveFactor = RemainingMoveFactor;
         RemainingMoveFactor = RemainingMoveFactor - CurMoveFactor;
         NumCompressedBytes += CurMoveFactor+1;
         if (NumCompressedBytes>MaxCompressedSize)
            {
            fprintf (stderr, "ERROR (Gcompress): Not enough room for new map.\n");
            exit(1);
            }
         *(CompressedBitMap++) = CurMoveFactor-1;
         while (CurMoveFactor--)
            *(CompressedBitMap++) = *(InputBitMap++);
         }


      }  while (InputBitMap<InputPtrLimit && CompressedBitMap<CompressedBitMapLimit);

   *CompressedSize = NumCompressedBytes;
   }


void GR_Expand
(
BYTE *InputBitMap,
BYTE *ExpandedBitMap,
WORD2  InputSize,
WORD2 *ExpandedSize
)
   {
   WORD2 NumRemainingInputChar;
   WORD2 NumOutputChar;
   WORD2 MaxNumOutputChar;
   WORD2 RepeatCount;
   WORD2 RepeatByte;
   WORD2 MoveCount;


   NumRemainingInputChar = InputSize;
   NumOutputChar         = 0;
   MaxNumOutputChar      = *ExpandedSize;
   while (NumRemainingInputChar > 0)
      {
      /*  Repeating byte  */
      if (InputBitMap[0] & 0x80)
         {
         RepeatCount = (*(InputBitMap++) & 0x007f) + 1;
         RepeatByte  = *(InputBitMap++);
         NumOutputChar += RepeatCount;
         NumRemainingInputChar  -= 2;

         /*  Insure that output doesn't overflow allocated buffer  */
         if (NumOutputChar > MaxNumOutputChar)
            {
            printf ("ERROR:  Expanded map exceeds allocated buffer size.\n");
            exit  (1);
            }

         /*  Write repeated byte  */
         memset (ExpandedBitMap, RepeatByte, RepeatCount);
         ExpandedBitMap += RepeatCount;

         }  /*  Repeat byte  */

      /*  Copy bytes from input map to output  */
      else
         {
         MoveCount = *(InputBitMap++) +1;
         NumOutputChar += MoveCount;
         NumRemainingInputChar  -= 1 + MoveCount;

         /*  Insure that output doesn't overflow allocated buffer  */
         if (NumOutputChar > MaxNumOutputChar)
            {
            printf ("ERROR:  Expanded map exceeds allocated buffer size.\n");
            exit  (1);
            }

         memmove (ExpandedBitMap, InputBitMap, MoveCount);
         ExpandedBitMap += MoveCount;
         InputBitMap    += MoveCount;


         }   /*  Copy bytes  */

      }  /*  while NumRemaingInputChar > 0  */

   *ExpandedSize = NumOutputChar;
   }

void GR_WriteCompressedBitmap (FILE *BitmapFile)
   {
   BYTE  *Bitmap;
   BYTE  *CompressedBitmap;
   WORD2  SizeBitmap;
   WORD2  SizeCompressedBitmap;
   ULONG NumDisplayBytesLeft;
   ULONG DisplayOffset;
   WORD2 TempSize;
   WORD2  OptimumBitmapSize;
   WORD2  OptimumCompressed;
   WORD2  Difference;


   /*  Get optimum bit map size for copying display buffer  */
   OptimumBitmapSize = GR_GetOptimumBufferSize();
   Difference        = 0xfff0 - OptimumBitmapSize;
   if (Difference > 0x1000)
      OptimumCompressed = OptimumBitmapSize + 0x1000;
   else
      OptimumCompressed = 0xfff0;

   /*
   Allocate bit maps used to copying display
   (need not fit entire display in one move)
   */
   ALLOCATE (Bitmap,           BYTE, OptimumBitmapSize);
   ALLOCATE (CompressedBitmap, BYTE, OptimumCompressed);

   /*
   Start with all display bytes still left
   and display offset a beginning of display
   */
   NumDisplayBytesLeft = GR_GetDisplaySize();
   DisplayOffset       = 0L;

   /*
   Carry bytes back from display in buffer until all of display read
   */
   while (NumDisplayBytesLeft > 0)
      {

      /*  Read some of display memory  */
      SizeBitmap = OptimumBitmapSize;;
      if (SizeBitmap > NumDisplayBytesLeft)
         SizeBitmap = (WORD2) NumDisplayBytesLeft;
      GR_ReadBuffer (Bitmap, DisplayOffset, &SizeBitmap);

#ifdef DEBUG
   if (SizeBitmap > OptimumBitmapSize)
      {
      GR_SetTextMode();
      printf ("INTERNAL ERROR: BITMAP SIZE.\n");
      exit   (1);
      }
#endif

      /*  Compress current portion of display bitmap  */
      SizeCompressedBitmap = OptimumCompressed;
      GR_Compress (Bitmap, CompressedBitmap, SizeBitmap, &SizeCompressedBitmap);

      /*  Write compressed bit map size  */
      TempSize = SizeCompressedBitmap;
      if (!GR_IsBigEndian())
         GR_ReverseEndian (&TempSize, sizeof(TempSize) );
      fwrite (&TempSize, sizeof(WORD2), 1, BitmapFile);

      /*  Write original bit map size  */
      TempSize = SizeBitmap;
      if (!GR_IsBigEndian())
         GR_ReverseEndian (&TempSize, sizeof(TempSize) );
      fwrite (&TempSize, sizeof(WORD2), 1, BitmapFile);

      /*  Write out compressed bit map  */
      fwrite
         (CompressedBitmap, sizeof(BYTE), SizeCompressedBitmap, BitmapFile);

      /*
      Increment/Decrement Display offset and number of display bytes left
      */
#ifdef DEBUG
   if (SizeBitmap > NumDisplayBytesLeft)
      {
      GR_SetTextMode();
      printf ("INTERNAL ERROR: NumDispBytesLeft too small.\n");
      exit   (1);
      }
#endif
      DisplayOffset += SizeBitmap;
      NumDisplayBytesLeft -= SizeBitmap;
      }


   /*  Free allocated memory  */
   FREE (CompressedBitmap);
   FREE (Bitmap);
   }



/*
Determine if current machine is big_endian
*/
BOOLEAN GR_IsBigEndian (void)
   {
   int i=1;
   char *c= (char *) &i;
   return (*c==1);
   }


/*
GR_ReverseEndian byte order for n bytes
*/
void GR_ReverseEndian (void *BufferPtr, int NumByte)
   {
   char *CharPtr = (char *) BufferPtr;
   char TempChar;
   int  FirstByte;
   int  LastByte;

   /*  Initialize pointer to first and last byte  */
   FirstByte = -1;
   LastByte  = NumByte;

   /*  Work in from ends, swapping bytes as you go  */
   while (++FirstByte<--LastByte )
      {
      TempChar           = CharPtr[FirstByte];
      CharPtr[FirstByte] = CharPtr[LastByte ];
      CharPtr[LastByte ] = TempChar;
      }

   }
