/*
************************************************************************
History
************************************************************************
*/
/*
11 March 1997
   GR_WritePixel(x,y,color)
     Have routine only write if x,y are within limits 
	  (0,0) (HorizontalResolution_g, VerticalResolution_g)
     This fixes (untested) problem when TAIL option to WMOVIEC too big.
*/
/*
Low Level Graphics Routines for EGA Card
*/

/*
Exported routines
*/
/*
void  GR_OpenGraphics       (WORD2 GraphicsMode);
void  GR_CloseGraphics      (void);
void  GR_SetGraphics        (void);
void  GR_SetTextMode        (void);
void  GR_GetResolution      (WORD2 *MaxX, WORD2 *MaxY, WORD2 *MaxColor);
void  GR_SetDisplayPage     (WORD2 Page);
void  GR_SetDrawPage        (WORD2 Page);
void  GR_ClearPage          (void);
void  GR_WritePixel         (WORD2 x, WORD2 y, WORD2 Color);
WORD2 GR_ReadPixel          (WORD2 x, WORD2 y);
ULONG GR_GetDisplaySize     (void);
void  GR_ReadBuffer         (BYTE *Buffer, ULONG Offset, WORD2 Size);
void  GR_WriteBuffer        (BYTE *Buffer, ULONG Offset, WORD2 Size);
void  GR_SetBackgroundColor (WORD2 BackgroundColor);
*/


/*
************************************************************************
Compile Switches
************************************************************************
*/
/*  Allow compilation of 386 assembly instructions  */
#pragma inline

#define DEBUG


/*
************************************************************************
File Includes
************************************************************************
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "llgraph.h"


/*
************************************************************************
Defines
************************************************************************
*/
#define MAX_PAGES 2
#define DEFAULT_BACKGROUND_COLOR 0

#define NUM_BYTES_PER_LINE  80L
#define NUM_LINES          350L
#define NUM_PLANES           4L

/*  Best buffer size for storing portion of display  */
#define OPTIMUM_BUFFER_SIZE  28000

/*
************************************************************************
Type Definitions
************************************************************************
*/



/*
************************************************************************
External (Global) Variables
************************************************************************
*/
WORD2 HorizontalResolution_g;
WORD2 VerticalResolution_g;
WORD2 BackgroundColor_g = DEFAULT_BACKGROUND_COLOR;



/*
************************************************************************
Module-wide variables
************************************************************************
*/
#ifdef GRAPHICS
static char  far *Display_m = NULL;
#else
static char  **Display_m;
#endif
static WORD2  NumDisplayPages_m;
static WORD2  NumPlanesUsed_m;
static WORD2  CurPage_m;
static ULONG NumCharsPerDisplay_m;
static ULONG NumCharsPerPlane_m;
/*  Record graphics state  */
static BOOLEAN OpenWasCalled_m     = FALSE;
static BOOLEAN GraphicsWasCalled_m = FALSE;


/*
************************************************************************
Local Function Prototypes
************************************************************************
*/
void LL_SetEGAWriteMode (WORD2);

/*
************************************************************************
Exported Routines
************************************************************************
*/

/*  Dummy routine, not required for EGA card  */
BOOLEAN GR_OpenGraphics (void)
   {


   /*  Set state  */
   OpenWasCalled_m = TRUE;

   /*  Return FALSE if no error  */
   return FALSE;
   }

/*  Dummy routine, not required for EGA card  */
void GR_CloseGraphics (void)
   {

   /*  Set State  */
   OpenWasCalled_m = FALSE;
   }


void GR_SetGraphicsMode (WORD2 NumColors)
   {
   WORD2 TempColors;
   int  iplane;

   /*  Set state  */
   GraphicsWasCalled_m = TRUE;

   /*
   Set horizontal resolution in long variable
   */
   HorizontalResolution_g = 640;
   VerticalResolution_g   = 350;

   /*  Test for too few colors  */
   if (NumColors < 2)
      {
      printf
         (
         "ERROR (Gr_SetGraphicsMode):  Too few colors requested (%u).  Must be two or more.\n",
         NumColors
         );
      exit (1);
      }

   /*  Determine number of planes used  */
   NumPlanesUsed_m = 0;
   TempColors      = NumColors-1;
   while (TempColors>0)
      {
      TempColors >>= 1;
      NumPlanesUsed_m ++;
      }

   /*  Test that number of planes does not exceed maximum  */
   if (NumPlanesUsed_m > NUM_PLANES)
      {
      printf ("ERROR (GR_SetGraphicsMode):  Too many colors requested (%u).\n",
         NumColors);
      printf ("Maximum allowed is 16.\n");
      exit   (1);
      }

   /*  Determine windows needed for clearing display pages  */
   NumCharsPerPlane_m       = NUM_BYTES_PER_LINE * NUM_LINES;
   NumCharsPerDisplay_m     = NumCharsPerPlane_m * NumPlanesUsed_m;

   /*  Enforce a maximum of MAX_PAGES display pages  */
   NumDisplayPages_m   = 2;
   if (NumDisplayPages_m > MAX_PAGES)
      NumDisplayPages_m = MAX_PAGES;

   /*  Set DRAW and DISPLAY pages  */
   GR_SetDrawPage    (0);
   GR_SetDisplayPage (0);

#ifdef GRAPHICS
   /*  Set EGA Draw Mode 2  */
   LL_SetEGAWriteMode (2);
#endif

   /*  Clear graaphics display (requires write mode is set to 2)  */
   CurPage_m = 0;

   /*  Allocate graphics storage  */
   Display_m = (char **) calloc (NumPlanesUsed_m, sizeof(char *));
   if (Display_m!=NULL)
      LOOP (iplane, NumPlanesUsed_m)
         Display_m[iplane] =
            (char *) calloc (NumCharsPerPlane_m, sizeof(char));

   if (Display_m==NULL || Display_m[NumPlanesUsed_m-1]==NULL)
      {
      printf ("ERROR: Cannot allocate memory for bit map storage.\n");
      exit   (1);
      }

   GR_ClearPage ();

   }

void GR_SetTextMode(void)
   {
	int iplane;

   /*  Set state  */
   GraphicsWasCalled_m = FALSE;

   /*  If no graphics memory then return  */
   if (Display_m==NULL)
      return;

   /*  Free graphics memory before return  */
   LOOP (iplane, NumPlanesUsed_m)
      if (Display_m[iplane]!=NULL)
         {
         free (Display_m[iplane]);
         Display_m[iplane] = NULL;
         }
   free (Display_m);
   Display_m = NULL;


   /*  Reset standard EGA Draw Mode  */
#ifdef GRAPHICS
   LL_SetEGAWriteMode (0);

   /*  Enter Text Mode  */
   asm{
      mov ax, 0x0003
      int 0x10
      }
#endif
   }


void GR_GetResolution (WORD2 *MaxX, WORD2 *MaxY, WORD2 *NumColorBits)
   {
   /*  Test for graphics mode called  */
   if (!GraphicsWasCalled_m)
      {
      GR_SetTextMode();
      printf ("INTERNAL ERROR:  Call to GR_GetResolution() without.\n");
      printf ("prior call to GR_SetGraphicsMode().\n");
      exit   (1);
      }

   *MaxX = HorizontalResolution_g;
   *MaxY = VerticalResolution_g;
   *NumColorBits = 4;
   }


#pragma argsused
void GR_SetDisplayPage (WORD2 Page)
   {
#ifdef GRAPHICS
   asm{
      mov ax, Page
      and al, 0x01
      mov ah, 0x05
      int 0x10
      }
#endif
   }


#pragma argsused
void  GR_SetDrawPage (WORD2 Page)
   {
#ifdef GRAPHICS
   Page &= 0x01;
   if (Page==0)
      Display_m = (char far *) 0xA0000000L;
   else
      Display_m = (char far *) 0xA8000000L;
#endif
   }


/*
Set graphics page to background color
(Assumes Write mode 2 already selected)
*/
void GR_ClearPage (void)
   {
   int iplane;
   int Color;

#ifdef GRAPHICS
   /*  Store bit mask 0xff  */
   asm{
      mov dx, 0x03ce;
      mov ax, 0xff08;
      out dx, ax;
      }
#endif

   /*  Fill display with background color  */
   Color = BackgroundColor_g;
   LOOP (iplane, NumPlanesUsed_m)
      {

      /*  Fill or clear plane according to bit value  */
      if ( (Color&0x01) > 0)
         memset (Display_m[iplane], 0xFF, (WORD2) NumCharsPerPlane_m);
      else
         memset (Display_m[iplane], 0x00, (WORD2) NumCharsPerPlane_m);

      /*  Set color for next plane  */
      Color >>= 1;
      }
   }


void GR_WritePixel (WORD2 x, WORD2 y, WORD2 Color)
   {
   /*  C version of pixel writing routine  */
   int iplane;
   WORD2  Offset;
   BYTE  Mask;


	/*  Reject point if out of picture  */
	if (x >= HorizontalResolution_g)
		return;
	if (y >= VerticalResolution_g)
		return;

   Offset  = y * 80 + (x >> 3);
   Mask    = 0x80 >> (x & 0x0007);

   /*  Write color to selected bit mask  */
   LOOP (iplane, NumPlanesUsed_m)
      {

      /*  If color bit on for this plane, set bit  */
      if (Color & 0x01)
         Display_m[iplane][Offset] |= Mask;

      /*  .. else clear bit  */
      else
         Display_m[iplane][Offset] &= (~Mask);

      Color >>= 1;
      }

   }


BYTE GR_ReadPixel (WORD2 x, WORD2 y)
   {
   WORD2  Offset;
   BYTE  Mask;
   BYTE  OutputColor;
   BYTE  iplane;
	BYTE  MaskedByte;

   Offset = y * 80 + (x >> 3);
   Mask   = 0x80 >> (x & 0x0007);

   /*  For each plane, read byte at offset  */
   OutputColor = 0;
   LOOP (iplane, NUM_PLANES)
      {

      /*  Shift previous value  */
      OutputColor <<= 1;

      /*  Read byte     */
      MaskedByte = (Display_m[iplane][Offset] & Mask);

      /*  If bit set, increment Outputcolor  */
      if (MaskedByte!=0)
         OutputColor++;
      }


   /*  Read color for selected bit  */
   return OutputColor;

   }


ULONG GR_GetDisplaySize (void)
   {
   /*  Test for graphics mode called  */
   if (!GraphicsWasCalled_m)
      {
      GR_SetTextMode();
      printf ("INTERNAL ERROR:  Call to GR_GetDisplaySize() without.\n");
      printf ("prior call to GR_SetGraphicsMode().\n");
      exit   (1);
      }

   return NumCharsPerDisplay_m;
   }


/*
Read *Size bytes of display buffer into Buffer[] starting at offset.
The "display buffer" is all 4 planes of the EGA memory, and is treated
as contiguous memory indexed from 0 to NumCharsPerDisplay_m-1

BYTE  *Buffer     - buffer to contain data from display
ULONG FirstOffset - starting position in display
WORD2  *Size       - Number of bytes to read, returns number actually read
*/
void  GR_ReadBuffer (BYTE *Buffer, ULONG FirstOffset, WORD2 *Size)
   {
   ULONG LastOffset;
   WORD2  FirstPlane;
   WORD2  LastPlane;
   BYTE  iplane;
   ULONG CurFirstOffset;
   ULONG CurLastOffset;
   ULONG PlaneOffset;
   WORD2  OffsetInPlane;
   WORD2  SizeInPlane;
   WORD2  BufferOffset;


   /*  Find last byte to read  */
   LastOffset = FirstOffset + *Size;

   /*  If last offset beyond end of buffer, reset last offset and size  */
   if (LastOffset > NumCharsPerDisplay_m)
      {
      LastOffset = NumCharsPerDisplay_m;
      *Size      = (WORD2) (LastOffset - FirstOffset);
      }

   /*  Calculate first and last plane  */
   FirstPlane =  (WORD2) (  FirstOffset   / NumCharsPerPlane_m );
   LastPlane  =  (WORD2) ( (LastOffset-1) / NumCharsPerPlane_m );

   /*  Loop over planes to get bytes  */
   BufferOffset = 0;
   PlaneOffset  = FirstPlane * NumCharsPerPlane_m;
   for (iplane=FirstPlane; iplane<=LastPlane; iplane++)
      {

      /*  Find first and last offset relative to entire display space  */
      CurFirstOffset = PlaneOffset;
      CurLastOffset  = PlaneOffset + NumCharsPerPlane_m;

      /*  Adjust these offsets if the exceed requested offsets  */
      if (CurFirstOffset < FirstOffset)
         CurFirstOffset = FirstOffset;
      if (CurLastOffset  > LastOffset )
         CurLastOffset  = LastOffset;

      /*  Convert to first and last offset relative to start of current plane  */
      OffsetInPlane = (WORD2) (CurFirstOffset - PlaneOffset   );
      SizeInPlane   = (WORD2) (CurLastOffset  - CurFirstOffset);

#ifdef GRAPHICS
      /*  Select current read plane  */
      asm{
         mov ah, iplane
         mov al, 0x04
         mov dx, 0x03ce
         out dx, ax
         }
#endif


      /*  Move display memory to input buffer  */
#if 0
      movmem (Display_m[iplane] + OffsetInPlane, Buffer + BufferOffset, SizeInPlane);
#endif
      memmove (Buffer + BufferOffset, Display_m[iplane] + OffsetInPlane, SizeInPlane);

      /*  Increment offset into Display buffer and input buffer  */
      PlaneOffset  += NumCharsPerPlane_m;
      BufferOffset += SizeInPlane;


      }

#ifdef DEBUG
      if (BufferOffset != *Size)
         {
         GR_SetTextMode();
         printf ("ERROR (GR_ReadBuffer):  Incorrect number of bytes written.\n");
         exit   (1);
         }
#endif

   }



/*
Write *Size bytes into display buffer from Buffer[] starting at offset.
The "display buffer" is all 4 planes of the EGA memory, and is treated
as contiguous memory indexed from 0 to NumCharsPerDisplay_m-1

BYTE  *Buffer     - Buffer to contain data to write
ULONG FirstOffset - Starting position in display
WORD2  *Size       - Number of bytes to write, returns number actually written
*/
void  GR_WriteBuffer (BYTE *Buffer, ULONG FirstOffset, WORD2 *Size)
   {
   ULONG LastOffset;
   WORD2  FirstPlane;
   WORD2  LastPlane;
   BYTE  iplane;
   ULONG CurFirstOffset;
   ULONG CurLastOffset;
   ULONG PlaneOffset;
   WORD2  OffsetInPlane;
   WORD2  SizeInPlane;
   WORD2  BufferOffset;


   /*  Find last byte to read  */
   LastOffset = FirstOffset + *Size;

   /*  If last offset beyond end of buffer, reset last offset and size  */
   if (LastOffset > NumCharsPerDisplay_m)
      {
      LastOffset = NumCharsPerDisplay_m;
      *Size      = (WORD2) (LastOffset - FirstOffset);
      }

   /*  Calculate first and last plane  */
   FirstPlane = (WORD2) ( FirstOffset    / NumCharsPerPlane_m );
   LastPlane  = (WORD2) ( (LastOffset-1) / NumCharsPerPlane_m );

   /*  Set write mode 0  */
   LL_SetEGAWriteMode (0);

   /*  Set bit mask to entire byte  */
#ifdef GRAPHICS
   asm{
      mov dx, 0x03ce
      mov ax, 0xff08
      out dx, ax
      }
#endif


   /*  Loop over planes to get bytes  */
   BufferOffset = 0;
   PlaneOffset  = FirstPlane * NumCharsPerPlane_m;
   for (iplane=FirstPlane; iplane<=LastPlane; iplane++)
      {

      /*  Find first and last offset relative to entire display space  */
      CurFirstOffset = PlaneOffset;
      CurLastOffset  = PlaneOffset + NumCharsPerPlane_m;

      /*  Adjust these offsets if the exceed requested offsets  */
      if (CurFirstOffset < FirstOffset)
         CurFirstOffset = FirstOffset;
      if (CurLastOffset  > LastOffset )
         CurLastOffset  = LastOffset;

      /*  Convert to first and last offset relative to start of current plane  */
      OffsetInPlane = (WORD2) (CurFirstOffset - PlaneOffset   );
      SizeInPlane   = (WORD2) (CurLastOffset  - CurFirstOffset);

#ifdef GRAPHICS
      /*  Select current write plane  */
      asm{

         /*  Form mask for current plane  */
         mov cl, iplane
         and cl, 0x07
         mov ah, 0x01
         shl ah, cl

         /*  Map Mask Register Index  */
         mov al, 0x02

         /*  Write mask to register  */
         mov dx, 0x03c4
         out dx, ax
         }
#endif


      /*  Move display memory to input buffer  */
      /*  NOTE:  May have to "latch" byte first  */
#if 0
      movmem (Buffer + BufferOffset, Display_m[iplane] + OffsetInPlane, SizeInPlane);
#endif
      memmove (Display_m[iplane] + OffsetInPlane, Buffer + BufferOffset, SizeInPlane);

      /*  Increment offset into Display buffer and input buffer  */
      PlaneOffset  += NumCharsPerPlane_m;
      BufferOffset += SizeInPlane;


      }

   /*  Reset write mode 2  */
   LL_SetEGAWriteMode (2);

#ifdef DEBUG
      if (BufferOffset != *Size)
         {
         GR_SetTextMode();
         printf ("ERROR (GR_WriteBuffer):  Incorrect number of bytes written.\n");
         exit   (1);
         }
#endif
   }



void GR_SetBackgroundColor (WORD2 Color)
   {
   BackgroundColor_g = Color;
   }

WORD2 GR_GetOptimumBufferSize (void)
   {
   return OPTIMUM_BUFFER_SIZE;
   }

/*
************************************************************************
Local graphics subroutines
************************************************************************
*/

void LL_SetEGAWriteMode (WORD2 WriteMode)
   {
#ifdef GRAPHICS
   asm{
      /*  SET WRITE MODE  */
      mov dx, 0x03ce
      mov al, 0x05
      out dx, al
      mov ax, WriteMode
      inc dx
      out dx, al
      }
#endif
   }
