/*
***********************************************************************
Modification History
***********************************************************************
*/
/*
13 Dec 1991:  CurStep-sel was not allocated (because readrcv does not allocate)
              inserted explicit allocation in main loop.
 3 Jan 1992:  built in qsort was bombing the program, perhaps because of
              stack overflow(?).  Substituted Shell sort that does not
              use stack recursively.
20 Feb 1992:  Error in BONDSUB.C.  Loop over j=(i+1)%3 was previously done
              without  %3.
20 Apr 1992:  In DRAWSTEP, was not passing CurStep->boundrep to BONDSUB.
 8 Mar 1993:  (1) Draw bonds BEFORE particles
              (2) Allow users to break WMOVIE routine
26 Apr 1993:  Add tails to movies
 7 Oct 1994:  - Added FAR define that evaluates to far under PC,
                nothing under UNIX
              - Changed call to BORLAND proprietary movmem() to
                standard UNIX memmove().
              - Removed // comments from header files.
              - Remove includes to Borland header files alloc.h and mem.h
                (substitute string.h for mem.h)
              - For some reason wmovie.c will not compile on HOLLY with
                casting strdup() to (char *) strdup().  However, a simple
                test program with strdup() does not require this cast.
                Strange.
              - Because Borland and AIX integers differ in size (2 and 4
                bytes respectively), must restrict the fwrite() statement
                to write out only 2 bytes (otherwise AIX version will
                write out 4 bytes).
19 Oct 1994:  - On DEC Alpha program gave memory fault, because while{}
                loop would access argv[1][0] when argv[1] was NULL.
                Have added a test to while condition to insure there is
                is sufficient number of elements in argv[].
1 Feb 1995:   - Version was not using types correctly.  Have cleaned
                up code in drawstep.c and renamed variables, and found
                that was not using correct indices for particles in
                for (isort..) loop.  Was using isort where should have3
                used image.
              - Also, removed all calls to strdup().
2 Mar 1995    - Removed call to PSUB, used SelectLayer() function instead.
              - Added -RefStep option.
VERSION 1.2.0
12 Jun 1995   - Added option to plot tails relative to previous step (-m)
VERSION 1.2.1
12 Jun 1995   - removed unneeded line-feed from Usage_m[] documentation
              - changed order of -d and -s options, made scale first option
              - added two of three current "sanity checks".
              - changed meaning of -s option, now modifies -m and -d options.
13 Jun 1995   - Add option to suppress atom symbols
21 Jun 1995   - Convert program to color EGA
              - eliminate code for image file
              - new format for movie file header
28 Jun 1995   - implemented command file to move options from the command
                line to a file.  Still accepts command line arguments.
 5 Jul 1995   - in Drawstep replaced calloc() and free() with
                ALLOCATE() and FREE()
              - and added module variables fir SlipCallBack() and
                NormalizeSlipDisplacements()
              - and moved local functions to bottom of file.
 6 Jul 1995   - added RCV file and MOV file names to command line, when
                using a command file.
              - corrected vectdisp command.
              - fixed memory allocation error was not freeing allocated
                memory for displacement of atoms. (in drawstep)
25 Jul 1995   - corrected tabbing and placements of brackets {,}
              - added more comments and white space
27 Jul 1995   - changed code for interpretation of command file to use
                strsub.c and strsub.h functions, which simplifies the code.

30 Aug 1995 - Don Macnamara has made extensive changes include palette
              information.

30 Aug 95   - Change variable magnitude from int to float.
            - Added call to ShutdownGraphics after check for
              magnitude equal to 0.  This prevents program from
              exiting in middle of graphics mode.
31 Aug 95   - Add RGB format for specifying colors from 64 palette
14 Dec 95   - Add option to color atoms by layer (for Pei's vacancy
              diffusion).
 9 Apr 96   - Corrected code to append movies.  New movie section does
              not include palette now.
 3 May 96   - Correct calculation of displacement in DS_drawstep()
              (see note in drawstep.c)
10 May 96   - More corrections to calculation of displacement.
              Was using wrong index to CoordCurHor[], ..Ver[], etc
              when using them to calculate DispHor, DispVer, ...
17 May 96   - Add options for coloring by "inplane" displacement
              magnitude.
VERSION 1.5.3
10 Jul 96   - Add VIEW option to change viewpoint
            - Add FRAME option to keep constant scale
Version 1.5.4
25 Nov 96   - COR and RCV routines returned coordinates in
              different units (cm and ang respectively).
              Set both to return cm, and changed input parameters
              to cm.
VERSION 1.5.5
20 Dec 96   - Correction to SelectLayer().  Sometime, due to roundoff
              error, the layer number for an atom would equal
              the number of layers, instead of ranging between 0
              and NumLayer-1.  Corrected this by adding an if 
              statement that mapped ilayer to NumLayer-1 if it was
              originally equal to NumLayer.
VERSION 1.5.6
 2 Jan 97   - Correction to ORIENT command as read from CMD file.
              Was not setting value of PlotParam->Rotate[] array
              according to the value of orient command.
VERSION 1.5.7
 9 Jan 97   - Suprisingly the RISC 6000 version of WMOVIEC did
              not correct NumPalette_m when writing binary file
              to BigEndian.  This needed to be done in wmoviec.c.
              Only in hlgraph.c was endian info taken care of.

VERSION 1.5.8
 5 Mar 97   - When reading in TYPELIST or other file names the
              trailing \n was stored along with name.  Created
              routine called TruncateString() to remove \n.

VERSION 1.5.9
 6 Mar 97    Previously, the TAIL command without an argument
             scaled the tails by 0.  Now it scales it by 1.

VERSION 1.5.10
 11 Mar 97  - Tail command with too big a scale would bomb
              program.  Problem appears to have been that
              tail ran off the display, and the routine
              GR_WritePixel(x,y,c) (in llmem.c) routine 
              did not clip.  Have changed the GR_WritePixel()
              routine to make it clip.
            - Frame command previosly did not clip particles
              outside frame.  It should do it now. 
VERSION 1.5.11
 20 Mar 97  - New clip routine ClipPoint() had mistake in it.
              It was comparing y coordinate with minimum x
              coordinate.
            - GR_DrawStep() routine had error, it required
              NumImagePart to equal number of selected particles.
              However no that particles were clipped inside frame,
              NumImagePart sometimes was not number of selected
              particles.  Changed comparison to check state
              of FixedFrame flag.

VERSION 1.5.12
 15 Apr 98  - Add SLIP to Usage documentation

VERSION 1.5.13
 15 Apr 98   - Change ATOM OFF so that tails are still drawn.
             - Add ATOM RADIUS TYPE n type1 radius type radius ..
VERSION 1.5.14
 25 Jan 2017 - In cor0.c, change INT2 from long to int, because
               in new compilers, int,long are 4,8 bytes.  IN
               original compiler, it was 2,4 bytes.
             - For fscan, fread, fscanf, assign return values
               to variables to avoid compiler warnings.
             - Functions memset() and memmove() are now in
               strings.h, so include it where necessary.
             - When orientation was 1 (i.e. y/1 instead of y/x),
               segmentation fault occurred.  Improved argument checking.
             - Fix year in date string in *.mov file
             - Code now compiles under GCC in Ubuntu 16.4 w/o errors.
*/

/*
***********************************************************************
Compile Switches
***********************************************************************
*/
#ifndef NOGRAPHICS
#ifdef __TURBOC__
#define GRAPHICS
#endif
#endif



/*
***********************************************************************
Include Files
***********************************************************************
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include <cor0.h>

#include "hlgraph.h"
#include "llgraph.h"
#include "drawstep.h"
#include "strsub.h"

#ifdef __TURBOC__
#include <conio.h>
#endif

/*
***********************************************************************
Defines
***********************************************************************
*/

/*  Default atom color  */
#define DEF_ATOM_COLOR 0

/*  Default aspect ratio  */
#define DEF_ASPECT_RATIO 1.0

/*  Maximum number of colors in palette  */
#define MAX_PALETTE 16

/*  Define direction constants  */
#define X 0
#define Y 1
#define Z 2
#define NDIR 3

/*  Number of characters in default buffer  */
#define STR_LEN 256

/*  Number of lines on console  */
#define NSCREEN   20

/*  Some ASCII Key Codes  */
#define CHAR_ESC 27
#define CHAR_SPACE 32
#define CHAR_ENTER 13

/*  Define EGA color mode  */
#define EGA

/*  Define PI  */
#ifndef PI
#define PI 3.14159265358979323846
#endif


/*  If PC use FAR modifier  */
#ifdef __TURBOC__
#define FAR far
#else
#define FAR
#endif



/*
***********************************************************************
Macros
***********************************************************************
*/

#define CROSS(C,A,B) \
   { \
   C[X]  =  A[Y] * B[Z] - A[Z] * B[Y]; \
   C[Y]  =  A[Z] * B[X] - A[X] * B[Z]; \
   C[Z]  =  A[X] * B[Y] - A[Y] * B[X]; \
   }

#define NORMALIZE(V) \
   { \
   Mag = (V)[X]*(V)[X] + (V)[Y]*(V)[Y] + (V)[Z]*(V)[Z]; \
   Mag = sqrt(Mag); \
   (V)[X] /= Mag; \
   (V)[Y] /= Mag; \
   (V)[Z] /= Mag; \
   }

#define SWAP(A,B,T) \
   if (A>B) \
      { \
      T = A; \
      A = B; \
      B = T; \
      }



/*  Write out current file name and line number  */
#define WRITEMSG \
   printf ("In file %s at line %i.\n", __FILE__, __LINE__);

/*  Write out variable name and value  */
#define WRITEVAR(VAR,TYPE) \
   printf ("File %s  Line %i : ", __FILE__, __LINE__); \
   printf ("%s = ", #VAR); \
   printf (#TYPE, VAR); \
   printf ("\n");


#define ALLOCATE(VAR,TYPE,NUM) \
   { \
   VAR = (TYPE *) calloc ( (NUM), sizeof ( TYPE ) ); \
   if (VAR == NULL) \
      { \
      printf ("ERROR (File %s line %i)\n", __FILE__, __LINE__); \
      printf ("   Cannot allocate variable %s.\n", #VAR); \
      ShutdownProgram(); \
      } \
   }

#define FREE(VAR) \
   { \
   if ((VAR)!=NULL) \
      { \
      free (VAR); \
      (VAR) = NULL; \
      } \
   }

/*  Simple loop (this macro help reduce typos)  */
#define LOOP(INDEX,LIMIT) \
   for (INDEX=0; INDEX<(LIMIT); INDEX++)



/*
***********************************************************************
Module Wide Variables
***********************************************************************
*/

/*  Set if program error occured  */
BOOLEAN ErrorStatus_m;

/*  Color palette information  */
WORD2    NumPalette_m = 0;
BYTE     PaletteColor_m [MAX_PALETTE+1];

/*  Usage message  */
char *Usage_m[] =
{
"",
" WMOVIEC -?    Print command list",
"",
" WMOVIEC command_file data_file movie_file",
"  command_file  is the name of the command file.  The file SAMPLE.CMD is an",
"                example of a command file.",
"  data_file is the name of the RCV file.  This file is used as input to",
"                create the movie file.",
"  movie_file is the name of the output movie file name.  The program",
"                creates the movie with the name specified.",
"",
" or",
"",
" WMOVIEC [..options..]  frcv [+]fmov orient nlayer ilayer",
"    frcv      is input  RCV file",
" [+]fmov      is output movie file (+ indicates append)",
"    orient    is the orientation (for example Y/X or Z/X etc.)",
"               (first dimension runs vertical, second horizontal)",
"    nlayer    number of layers",
"    ilayer    layer plotted",
" OPTIONS",
"   [-t ftype]      TYPELIST file (first line TYPELIST n followed by types)",
"   [-x      ]      Display Particle Index Number rather than image.",
"   [-n      ]      Display no atom images (only tails and bonds).",
"   [-b b1 b2]      Draw bonds connecting atoms between b1 and b2 apart.",
"   [-w      ]      Wrap bonds around repeating boundary conditions.",
"   [-r RefFile ]      Use first step in reference file for selecting layer",
"   [-d RefFile scale] Draw tails to positions in first step of RCV file",
"                    REFFILE.  Scale tails by SCALE.",
"   [-m scale     ] Draw tails relative to previous step.  Scales by SCALE.",
"   [-s rad       ] Draws 'slip' tails.  Modifies -d and -m options.  RAD ",
"                    is radius of spherical neighborhood of each particle.",
" Writes movie file in fmov.  (Toggle Pause)/(Quit) with [SPACE]/[ESC].",
" Version 1.5.13   (Use VMOVIEC for viewing)",
__DATE__,
NULL
};


/*  List of commands  */
char *CommandList_m[] =
{
"",
"WMOVIE COMMANDS:",
"",
"  ATOM {ON | OFF | INDEX }",
"  ATOM COLOR  TYPE       n  type1 color1 type2 color2 ... typen colorn",
"  ATOM COLOR  DISP   MAG Mag                       n color1 color2 ... colorn",
"  ATOM COLOR  DISP  PMAG Mag                       n color1 color2 ... colorn",
"  ATOM COLOR  DISP  RMAG                           n color1 color2 ... colorn",
"  ATOM COLOR  DISP PRMAG                           n color1 color2 ... colorn",
"  ATOM COLOR  DISP  VECT Minvect Maxvect Vx Vy Vz  n color1 color2 ... colorn",
"  ATOM COLOR  DISP RVECT                 Vx Vy Vz  n color1 color2 ... colorn",
"  ATOM COLOR  DISP ANGLE MinMag Phase              n color1 color2 ... colorn",
"  ATOM COLOR  LAYER                                n color1 color2 ... colorn",
"  ATOM RADIUS TYPE       n  type1 color1 type2 color2 ... typen colorn",
"  BOND        [WRAP]             Lobond Hibond",
"  BOND        [WRAP] COLOR DIST  Lobond Hibond     n color1 color2 ... colorn",
"  DISP        [REF Reffile]  [SLIP sliprad]",
"  FRAME       MinX MinY MinZ  MaxX MaxY MaxZ",
"  LABEL       {ON | OFF}",
"  LAYER       [REF Reffile] Nlayer Ilayer",
"  RESOLUTION  Xres Yres",
"  TAIL        [Scale]",
"  TITLE       Ntitle",
"   ...",
"  TYPELIST    Typefile",
"  VIEW        ViewX ViewY ViewZ  VertX VertY VertZ",
NULL
};


/*
***********************************************************************
Local Prototypes
***********************************************************************
*/

#ifdef GRAPHICS
void SetPalette (BYTE *InputPalette);
#endif

void     SelectLayer (stepdata *, int, int, int);
void     ShutdownProgram (void);
void     ShutdownGraphics (void);
void     WriteStringList (char **StringList, BOOLEAN IsPauseAllowed);
char     *GetCurrentTimeStr(void);
void     CopyCoordBetweenSteps (stepdata *, stepdata *);
int      ColorFromStr (char *ColorStr);
void     AddColorToPalette (WORD2 Color, WORD2 *Index);
int      ToUpper (int);
void     ParseOrientStr (char *InputStr, int *hor, int *ver, int *dep);
void     ReadCommandFile (char *, dsparm *);
void     ReadCommandLine (int, char *[], dsparm *, WORD2 *);
void     InitializeDisplayParam (dsparm *DisplayParam);
void     Normalize (float Vector[NDIR]);
BOOLEAN  CompareStrings (char *FirstStr, char *SecondStr);
void     InitializeView (dsparm *, double[NDIR], double[NDIR]);
void     TruncateString (char *);

/*
***********************************************************************
Main Routine
***********************************************************************
*/

int main (int argc, char *argv[])
   {
   char *InputFileName;
   char *MovieFileName;
   char *orient;
   int  ititle;
   int  ipalette;
   WORD2 BackgroundColor;
   WORD2 FrameColor;
   WORD2 TextColor;
   WORD2 DispColor;
   WORD2 BondColor;
   FILE      *MovieFile;
   FILE      *TypeListFile;
   CFILE     *LayerFile;
   CFILE     *RefFile;
   CFILE     *InputFile;
   stepdata  *CurStep;
   stepdata  *RefStep;
   dsparm    *DisplayParam;
   WORD2      MaxX;
   WORD2      MaxY;
   WORD2      MaxColor;
   WORD2     NumTokensRead;
   WORD2     NumTokensLeft;
   int      ReadStatus;
   BOOLEAN  FirstStep;
   int      idir;
   int      jdir;
   int      Page = 0;
   char     *LineFeed = "\012\015";
	WORD2    OutputNumPalette;


   /*  If using PC, declare variable for keyboard input  */
#ifdef __TURBOC__
   int InputChar;
#endif

   /*  If no argument, print usage message  */
   if (argc==1)
      {
      WriteStringList (Usage_m, FALSE);
      return 0;
      }

   /*  Test for request for command list  */
   if (CompareStrings(argv[1], "-?"))
      {
      WriteStringList (CommandList_m, FALSE);
      return 0;
      }

   /* Initialize Color Palette  */
   AddColorToPalette (63, &BackgroundColor);
   AddColorToPalette (00, &FrameColor);
   AddColorToPalette (00, &TextColor);
   AddColorToPalette (00, &DispColor);
   AddColorToPalette (00, &BondColor);

   /*  Allocate storage for display parameters  */
   ALLOCATE (DisplayParam,      dsparm,     1     )

   /*  Initialize flags  */
   InitializeDisplayParam (DisplayParam);

   /*  First pass at reading options  */
   ReadCommandLine (argc, argv, DisplayParam, &NumTokensRead);

   /*  Find number of remaining tokens  */
   NumTokensLeft = argc - NumTokensRead - 1;

   /*  Three tokens, read command file  */
   if (NumTokensLeft==3)
      {
      ReadCommandFile (argv[NumTokensRead+1], DisplayParam);
      InputFileName =  argv[NumTokensRead+2];
      MovieFileName =  argv[NumTokensRead+3];
      }

   /*  Five options  */
   else if (NumTokensLeft==5)
      {
      InputFileName         =       argv [NumTokensRead+1];
      MovieFileName         =       argv [NumTokensRead+2];
      orient                =       argv [NumTokensRead+3];
      DisplayParam->nlayer  = atoi (argv [NumTokensRead+4]);
      DisplayParam->ilayer  = atoi (argv [NumTokensRead+5]);

      /*  DETERMINE ORIENTATION  */
      ParseOrientStr
         (
         orient,
         &DisplayParam->hor,
         &DisplayParam->ver,
         &DisplayParam->dep
         );

      /*  Set rotation matrix to zero  */
      LOOP (idir,NDIR)
      LOOP (jdir,NDIR)
         DisplayParam->Rotate[idir][jdir] = 0;

      /*  Set matrix according to orientation  */
      DisplayParam->Rotate[X][DisplayParam->hor] = 1.0;
      DisplayParam->Rotate[Y][DisplayParam->ver] = 1.0;
      DisplayParam->Rotate[Z][DisplayParam->dep] = 1.0;
      }

   /*  else unsupported number of remaining tokens, print message  */
   else
      {
      WriteStringList (Usage_m, FALSE);
      return 0;
      }

   /*  Re-read command switches (to over-ride command file)  */
   ReadCommandLine (argc, argv, DisplayParam, &NumTokensRead);

   /*  Normalize displacement vector  */
   if
   (
   DisplayParam->ColorAtomByDispVectFlag ||
   DisplayParam->ColorAtomByRelDispVectFlag
   )
      Normalize (DisplayParam->vector);



   /*
   *********************
   SCAN THROUGH RCV FILE
   *********************
   */

   /*  Allocate memory  */
   ALLOCATE (CurStep,           stepdata,   1     )
   RefStep = NULL;

   /*
   Read Layer Info from a Reference Step
   */
   if (DisplayParam->LayerRefFlag)
      {

      /*  Open file  */
      LayerFile = OpenCoordinateFile (DisplayParam->LayerRefFileName, "r");

      /*  Allocate space for coordinates info  */
      ALLOCATE (CurStep, stepdata, 1)

      /*  Read coordinates for reference step  */
      ReadCoordinateFile (LayerFile, CurStep);

      /*  Close reference file  */
      CloseCoordinateFile(LayerFile);

      /*  Allocate space for select array  */
      ALLOCATE (CurStep->sel, unsigned char, CurStep->np);

      /* Select layer  */
      SelectLayer
         (
         CurStep,
         DisplayParam->dep,
         DisplayParam->nlayer,
         DisplayParam->ilayer
         );

      /*  Set flag to preserve selection when reading new steps  */
      CurStep->selkeep = TRUE;

      /*  Free Reference step  */
      FreeStep (RefStep);

      }


   /*  READ TYPE FILE   */
   if (DisplayParam->TypeListFlag)
      {
      TypeListFile = fopen (DisplayParam->TypeListFileName, "rt");
      if (TypeListFile==NULL)
         {
         fprintf (stderr, "ERROR:  Cannot open type file %s.\n",
            DisplayParam->TypeListFileName);
         return 1;
         }
      readtype (TypeListFile, CurStep);
      fclose (TypeListFile);

      /*  Set flag to maintain type values  */
      CurStep->selkeep = TRUE;
      }


   /*  Append to existing movie file ...  */
   if (MovieFileName[0]=='+')
      MovieFile = fopen (MovieFileName+1,  "ab");

   /*  ... else Open and intialize new movie file  */
   else
      MovieFile = fopen (MovieFileName,    "wb");

   /*  Test for error opening file  */
   if (MovieFile==NULL)
      {
      fprintf (stderr, "ERROR:  Cannot open output file %s.\n", MovieFileName);
      return 1;
      }


   /*  If new file, write file header  */
   if (MovieFileName[0]!='+')
      {
      /*  Write movie id string  */
      fwrite (" EGA MOVIE FILE: Version 1.5.4", 30, 1, MovieFile);
      fwrite (LineFeed, strlen(LineFeed), 1, MovieFile);

      /*  Write date and time  */
      fwrite (" DATE: ", 7, 1, MovieFile);
      fwrite (GetCurrentTimeStr(), 21, 1, MovieFile);
      fwrite (LineFeed, strlen(LineFeed), 1, MovieFile);

      /*  Write RCV file name  */
      fwrite (" INPUT FILE: ", 13, 1, MovieFile);
      fwrite (InputFileName,   strlen(InputFileName),   1, MovieFile);
      fwrite (LineFeed, strlen(LineFeed), 1, MovieFile);

      /*  Write one leading blank for title  */
      fwrite (" ", 1, 1, MovieFile);

      /*  Movie file comments go here, at header end  */
      LOOP (ititle, DisplayParam->NumTitle)
         {
         fwrite
            (
            DisplayParam->Title[ititle],
            strlen(DisplayParam->Title[ititle]),
            1,
            MovieFile
            );
         fwrite (LineFeed, strlen(LineFeed), 1, MovieFile);
         }

      /*  Write header end (DOS End-of-file character)  */
      fwrite ("\032",   1,                1, MovieFile);

      /* Save Palette Information in movie file. */
		OutputNumPalette = NumPalette_m;
		if (!GR_IsBigEndian())
			GR_ReverseEndian (&OutputNumPalette, sizeof(OutputNumPalette));
      fwrite (&OutputNumPalette,  sizeof(OutputNumPalette), 1, MovieFile);
      LOOP (ipalette, NumPalette_m)
         fwrite
            (
            &PaletteColor_m[ipalette],
            sizeof(PaletteColor_m[ipalette]),
            1,
            MovieFile
            );

      }


   /*  Enter graphics mode  */
   GR_OpenGraphics   ( );
   GR_SetGraphicsMode(NumPalette_m);


   /*  Set color palette  */
#ifdef GRAPHICS
   SetPalette(PaletteColor_m);
#endif

   /*  Set colors  */
   GR_SetBackgroundColor (BackgroundColor);
   DS_SetFrameColor      (FrameColor);
   DS_SetTextColor       (TextColor);
   DS_SetDispColor       (DispColor);
   DS_SetBondColor       (BondColor);

   /*  Read resolution from graphics routines  */
   GR_GetResolution (&MaxX, &MaxY, &MaxColor);

   /*  Set resolution if not user specified  */
   if (DisplayParam->hsmax==0 || DisplayParam->vsmax==0)
      {
      DisplayParam->hsmax = MaxX-1;
      DisplayParam->vsmax = MaxY-1;
      }

   /*  Test for user resolution too big  */
   if (DisplayParam->hsmax>=MaxX || DisplayParam->vsmax>=MaxY)
      {
      ShutdownGraphics();
      printf ("ERROR:  User specified resolution (%i,%i)\n",
         DisplayParam->hsmax+1, DisplayParam->vsmax+1);
      printf ("exceeds graphics card maximum (%i,%i)\n",
         MaxX, MaxY);
      ShutdownProgram();
      }

   /*  Option for reading single reference step for all displacements  */
   if (DisplayParam->DispFlag && DisplayParam->DispRefFlag)
      {

      /*  Open file  */
      RefFile = OpenCoordinateFile (DisplayParam->DispRefFileName, "r");

      ALLOCATE (RefStep, stepdata, 1)

      ReadCoordinateFile (RefFile, RefStep);

      CloseCoordinateFile(RefFile);
      }




   /*
   *********************
   DRAW INDIVIDUAL STEPS
   *********************
   */


   ErrorStatus_m = FALSE;
   FirstStep     = TRUE;

   /*  Open input coordinate file  */
   InputFile = OpenCoordinateFile (InputFileName, "r");

   /*  Read first step from file  */
   ReadStatus = ReadCoordinateFile (InputFile, CurStep);

   while (ReadStatus!=EOF && (!ErrorStatus_m) )
      {

      /*  TEST FOR USER BREAK  */
#ifdef __TURBOC__
      if (kbhit())
         {
         InputChar = getch();

         /*  Leave program if ESC pressed  */
         if (InputChar==CHAR_ESC)
            break;

         /*  If SPACE then pause program until next SPACE  */
         if (InputChar==CHAR_SPACE)
            {
            do
               {
               InputChar = getch();

               /*  Leave program if ESC pressed  */
               if (InputChar==CHAR_ESC)
                  break;
               }
               while (InputChar != CHAR_SPACE &&  InputChar != CHAR_ENTER);
            }
         }
#endif


      /*
      Check that reference and current step have same atom count
      */
      if (
         DisplayParam->DispRefFlag ||
         (DisplayParam->DispFlag && !FirstStep)
         )
      if (RefStep->np != CurStep->np)
         {
         ShutdownGraphics();
         printf ("ERROR:  Number of atoms (%i) in reference step\n",
            RefStep->np);
         printf ("        is not equal to number of atoms (%i)\n",
            CurStep->np);
         printf ("        in current step (%li)\n",
            CurStep->step);
         ShutdownProgram();
         }

      /*
      If not using reference step for selecting layer,
      select layer for current step
      */
      if (!DisplayParam->LayerRefFlag)
         {
         /*  ALLOCATE SELECT FLAG  */
         ALLOCATE (CurStep->sel, unsigned char, CurStep->np);

         /* SELECT LAYER ON THE FLY  */
         SelectLayer (CurStep, DisplayParam->dep, DisplayParam->nlayer, DisplayParam->ilayer);

         }

#if 0
      /*  GET MIN AND MAX LIMITS  */
      LOOP (idir, NDIR)
         {
         DisplayParam->cmin[idir] = -CurStep->trans[idir];
         DisplayParam->cmax[idir] =  CurStep->box[idir] - CurStep->trans[idir];
         }
#endif


      /*  DRAW ATOMS    */

      /*  Switch pages  */
      Page = 1-Page;

      /*  Set new page as drawing page  */
      GR_SetDrawPage  (Page);

      /*  Clear new page  */
      GR_ClearPage     ();

      /*  Draw image  */
      ErrorStatus_m =
         DS_DrawStep (RefStep, CurStep, DisplayParam);

      /*  Draw labels  */
      if (DisplayParam->LabelFlag)
         DS_DrawLabel (CurStep, DisplayParam);

      /*  Display new page  */
      GR_SetDisplayPage(Page);

      /*  Write new page  */
      GR_WriteCompressedBitmap  (MovieFile);

      /*  FREE SELECT FLAG  */
      if (!DisplayParam->LayerRefFlag)
			{
         FREE (CurStep->sel)
			}

      /*
      If plotting tails relative to previous step
      save current step as reference step for next time around
      */
      if (DisplayParam->DispFlag && !DisplayParam->DispRefFlag)
         {

         /*  Allocate reference step if not done before  */
         if (FirstStep)
            ALLOCATE (RefStep, stepdata, 1);

         /*  Copy coordinates from CurStep to RefStep  */
         CopyCoordBetweenSteps (RefStep, CurStep);
         }


      /*  Unset first step  */
      FirstStep = FALSE;

      /*  Read input file for next step  */
      ReadStatus = ReadCoordinateFile (InputFile, CurStep);


      }   /*  End of Draw individual steps  */

   /*  CLOSE FILES  */
   CloseCoordinateFile (InputFile);
   fclose (MovieFile);


   /*  Call error routine if error occured  */
   if (ErrorStatus_m)
      ShutdownProgram ();

   /*  CLOSE GRAPHICS  */
   ShutdownGraphics();



   /*  PROGRAM END  */
   return 0;

   }

/*
***********************************************************************
Local Subroutines
***********************************************************************
*/


#ifdef GRAPHICS
void SetPalette (BYTE *InputPalette)
   {
   int  ichar;
   char Palette[17];

   /*  Copy 16 palette colors  */
   LOOP (ichar, 16)
      Palette[ichar] = InputPalette[ichar];

   /*  Set overscan color to black  */
   Palette[16] = 0;


   asm{
      mov ax, 0x1002
      mov dx, ss
      mov es, dx
      lea dx, Palette
      int 0x10
      }

   }
#endif  /* ifdef GRAPHICS  */



void SelectLayer
(
stepdata *StepRecord,
int dep,
int NumLayer,
int ilayer
)
   {
   int ipart;
   int DepthLayer;
   float ScaleFac;
   float *Depth;
   float MinDepth;
   float MaxDepth;

   /*  Change layer range from [1..n] to [0..n-1]  */
   ilayer--;

   /*  Get pointer to coordinates used for depth  */
   Depth = StepRecord->c[dep];

   /*  Determine min and max depth values  */
   if (StepRecord->boundrep[dep])
      {
      MinDepth = 0;
      ScaleFac = NumLayer / StepRecord->box[dep];
      }
   else
      {
      MinDepth = Depth[0];
      MaxDepth = Depth[0];
      for (ipart=1;ipart<StepRecord->np;ipart++)
         {
         if (MinDepth>Depth[ipart])
            MinDepth = Depth[ipart];
         else if (MaxDepth<Depth[ipart])
            MaxDepth = Depth[ipart];
         }

      /*  Determine scale factor  */
      if (MinDepth==MaxDepth)
         ScaleFac = 0.0;
      else
         ScaleFac = NumLayer / (MaxDepth - MinDepth);
      }

   /*  SET SELECT FLAGS  */
   if (ScaleFac==0.0)
      {
      for (ipart=0;ipart<StepRecord->np;ipart++)
         StepRecord->sel[ipart] = TRUE;
      StepRecord->nsel = StepRecord->np;
      }
   else
      {
      StepRecord->nsel = 0;
      for (ipart=0;ipart<StepRecord->np;ipart++)
         {

         DepthLayer = ScaleFac * (Depth[ipart]-MinDepth);

			/*  Insure against roundoff error  */
			if (DepthLayer == NumLayer)
				DepthLayer = NumLayer-1;

         if (DepthLayer==ilayer)
            {
            StepRecord->nsel++;
            StepRecord->sel[ipart] = TRUE;
            }
         else
				{
            StepRecord->sel[ipart] = FALSE;
				}
         }
      }
   }


void CopyCoordBetweenSteps (stepdata *TargetStep, stepdata *SourceStep)
   {
   int icoord;


   /*  If target step has different number of coordinates  */
   if (TargetStep->np != SourceStep->np)
      {
      FREE (TargetStep->c[X]);
      FREE (TargetStep->c[Y]);
      FREE (TargetStep->c[Z]);
      TargetStep->np = 0;
      }

   /*  If target step has no coordinates  */
   if (TargetStep->np==0)
      {
      ALLOCATE (TargetStep->c[X], float, SourceStep->np)
      ALLOCATE (TargetStep->c[Y], float, SourceStep->np)
      ALLOCATE (TargetStep->c[Z], float, SourceStep->np)
      }

   /*  Copy coordinates  */
   LOOP (icoord, SourceStep->np)
      {
      TargetStep->c[X][icoord] = SourceStep->c[X][icoord];
      TargetStep->c[Y][icoord] = SourceStep->c[Y][icoord];
      TargetStep->c[Z][icoord] = SourceStep->c[Z][icoord];
      }
   TargetStep->np = SourceStep->np;


   }

/*  Clean up after error  */
void ShutdownProgram (void)
   {
#ifdef GRAPHICS
   int KeyPress;

   /*  Let user hit enter before clearing graphics screen  */
   printf ("Hit ENTER to continue.\n");
   KeyPress = getc (stdin);
   /*  This statement is just to avoid compiler warning  */
   KeyPress++;
#endif

   /*  Exit program  */
   exit (1);
   }

/*  Clean up before exiting program  */
void ShutdownGraphics(void)
   {
   GR_SetTextMode();
   GR_CloseGraphics();
   }


/*  Return time and date string  */
char *GetCurrentTimeStr (void)
   {
   struct tm *timeptr;
   static char TimeString[23];
   time_t CurrentTime;

   /*  Get current time  (in seconds from 1900)  */
   time (&CurrentTime);

   /*  Convert to date, time  */
   timeptr = localtime (&CurrentTime);

   /*  Form string  */
   sprintf (TimeString, "%02d/%02d/%04d   %02d:%02d:%02d",
      timeptr->tm_mon+1, timeptr->tm_mday, 1900+timeptr->tm_year,
      timeptr->tm_hour, timeptr->tm_min, timeptr->tm_sec);

   /*  Return string  */
   return TimeString;
   }




/*
************************************************************************
Subroutine to write StringList (Such as program Usage info)
************************************************************************
*/



void WriteStringList (char **StringList, BOOLEAN IsPauseAllowed)
   {
   char  InputChar;
   int   LinesLeft;
   char  **StringListPointer = StringList;

   LinesLeft = NSCREEN;
   InputChar='0';

   /*  Print out lines one at a time  */
   while ((IsPauseAllowed || *StringListPointer) && InputChar!='q')
      {
      /*  PRINT CHARACTER - COUNT LINEFEEDS  */
      puts (*StringListPointer);
      StringListPointer++;
      LinesLeft--;
      /*  END OF TEXT  */
      if (IsPauseAllowed && *StringListPointer==NULL)
         {
         StringListPointer = StringList;
         LinesLeft = 0;  /*  SIGNAL TO WAIT FOR KEYPRESS  */
         }
      if (!LinesLeft)
         {
         printf ("\n(q/ret):");
         InputChar = getchar();
         printf ("\n\n\n");
         LinesLeft = NSCREEN;
         }
      }
   }



/*
Convert color from format RGB to EGA color mode.
  For example, colors can be specified as

    302  which means 3-Red 0-Green 2-Blue  ->  reddish-purple
    000        means 0-Red 0-Green 0-Blue  ->  black
    333              3     3       3       ->  white
    111              1     1       1       ->  dark gray
    222              2     2       2       ->  light gray
*/
int ColorFromStr (char *ColorStr)
   {
   int EGAColor;
   int EGAColorComponent;
   int ColorLevel;
   int icolor;

   /*  Test for "clear" keyword  */
   if (CompareStrings(ColorStr, "CLEAR"))
      return CLEAR_COLOR;

   /*  Interpret as RGB 4 level code  */
   if (strlen(ColorStr)==3)
      {
      EGAColor = 0;
      LOOP (icolor, 3)
         {
         ColorLevel = ColorStr[icolor] - '0';
         if (ColorLevel>3)
            ColorLevel=3;
         switch (ColorLevel)
            {
            case 0:
               EGAColorComponent = 0;
               break;
            case 1:
               EGAColorComponent = 8;
               break;
            case 2:
               EGAColorComponent = 1;
               break;
            case 3:
               EGAColorComponent = 9;
               break;
            }
         EGAColor = (EGAColor << 1) + EGAColorComponent;
         }

      /*  Return final color  */
      return EGAColor;
      }

   /*  else interpret as straight EGA color  */
   return atoi (ColorStr);
   }



/*
Add color to palette (if not already present)
Return index to color (through passed pointer)

NOTE: Special color value CLEAR_COLOR is not added to palette,
      and value CLEAR_COLOR is passed back through *OutputIndex
*/
void AddColorToPalette (WORD2 InputColor, WORD2 *OutputIndex)
   {
   int index;

   /*  Test for clear color  */
   if (InputColor==CLEAR_COLOR)
      {
      *OutputIndex = CLEAR_COLOR;
      return;
      }

   /*  Is input color already in palette?  */
   LOOP (index, NumPalette_m)
      {

      /*  If color already in palette, return its index  */
      if (PaletteColor_m[index]==InputColor)
         {
         *OutputIndex = index;
         return;
         }

      }

   /*  Need to add color, check first for enough room  */
   if (NumPalette_m==MAX_PALETTE)
      {
      ShutdownGraphics();
      printf ("ERROR:  Too many colors requested (maximum is 16).\n");
      ShutdownProgram();
      }

   /*  OutputIndex points to new color entry  */
   *OutputIndex = NumPalette_m;
   PaletteColor_m[NumPalette_m] = InputColor;

   /*  Increment number of colors in palette  */
   NumPalette_m++;
   }


int ToUpper (int InputChar)
   {
   if (InputChar>='a' && InputChar<='z')
      InputChar += 'A' - 'a';
   return InputChar;
   }





/*
************************************************************************
Read command from command file
  -  Receive name of command file in      *InputFileName
  -  Store results in structure           *DisplayParam
************************************************************************
*/
void ReadCommandFile
(
char   *InputFileName,
dsparm *DisplayParam
)
   {
   FILE *InputFile;
	int   idir;
	int   jdir;
   int   WarningCount;
   int   ibond;
   int   icolor;
   int   itype;
   int   ititle;
   int   NumType;
   int   TempType;
   int   StringLength;
   WORD2  ColorIndex;
   WORD2  Color;
   char  *ColorStr;
   char  CmdFileBuffer[STR_LEN];
   char  *TempPtr;
   char  *HeadPtr;
   char  *TailPtr;
   double ViewAxis[NDIR];
   double VertAxis[NDIR];
   double TempDouble;

   /*  Cannot find command file  */
   if (NULL==(InputFile=fopen(InputFileName,"r")))
      {
      ShutdownGraphics();
      printf("Error: Command Line File not found.\n");
      ShutdownProgram();
      }

   /*  Read command file  */
   WarningCount = 0;
   while (NULL!=fgets(CmdFileBuffer,STR_LEN,InputFile))
      {

      /*  Skip comments  */
      if (CmdFileBuffer[0]=='#')
         continue;

		/*  Remove trailing \n  */
		TruncateString(CmdFileBuffer);

      /*  Read command  */
      TailPtr = CmdFileBuffer;
      HeadPtr = strhed(&TailPtr);

      /*  Skip empty strings   */
      if      (CompareStrings(HeadPtr, ""))
         ;

      /* orientation of movie */
      else if (CompareStrings(HeadPtr,"ORIENT"))
         {
         ParseOrientStr
            (
            TailPtr,
            &DisplayParam->hor,
            &DisplayParam->ver,
            &DisplayParam->dep
            );

         /*  Set rotation matrix to zero  */
         LOOP (idir,NDIR)
         LOOP (jdir,NDIR)
            DisplayParam->Rotate[idir][jdir] = 0;
   
         /*  Set matrix according to orientation  */
         DisplayParam->Rotate[X][DisplayParam->hor] = 1.0;
         DisplayParam->Rotate[Y][DisplayParam->ver] = 1.0;
         DisplayParam->Rotate[Z][DisplayParam->dep] = 1.0;
         }

      /*  Control display of movie label  */
      else if (CompareStrings(HeadPtr,"LABEL"))
         {
         HeadPtr = strhed (&TailPtr);

         /*  Show labels  */
         if (CompareStrings(HeadPtr,"ON"))
            DisplayParam->LabelFlag = TRUE;

         /*  No labels  */
         else if (CompareStrings(HeadPtr, "OFF"))
            DisplayParam->LabelFlag = FALSE;

         /*  Unknown option  */
         else
            {
            ShutdownGraphics();
            printf ("ERROR:  Unknown option for LABEL command.\n");
            ShutdownProgram();
            }
         }

      /* which layer to show */
      else if (CompareStrings(HeadPtr,"LAYER"))
         {
         HeadPtr = strhed (&TailPtr);
         if (CompareStrings(HeadPtr,"REF"))
            {
            DisplayParam->LayerRefFlag = TRUE;
            HeadPtr = strhed (&TailPtr);
            ALLOCATE
               (DisplayParam->LayerRefFileName, char, strlen(HeadPtr)+1)
            strcpy  (DisplayParam->LayerRefFileName, HeadPtr);
            HeadPtr = strhed (&TailPtr);
            }
         DisplayParam->nlayer = intstr(&HeadPtr);
         DisplayParam->ilayer = intstr(&TailPtr);
         }


      /* typelist */
      else if (CompareStrings(HeadPtr,"TYPELIST"))
         {
         DisplayParam->TypeListFlag=TRUE;
         ALLOCATE(DisplayParam->TypeListFileName, char, strlen(TailPtr)+1);
         strcpy  (DisplayParam->TypeListFileName, TailPtr);
         }


      /* draw bonds? */
      else if (CompareStrings(HeadPtr,"BOND"))
         {
         DisplayParam->BondFlag = TRUE;

         /*  Test for wrap option  */
         HeadPtr = strhed (&TailPtr);
         if (CompareStrings(HeadPtr, "WRAP"))
            {
            DisplayParam->BondWrapFlag = TRUE;
            HeadPtr = strhed (&TailPtr);
            }


         /*  Test for color option  */
         else if (CompareStrings(HeadPtr, "COLOR"))
            {

            DisplayParam->ColorBondByDistFlag = TRUE;

            HeadPtr = strhed (&TailPtr);
            if (CompareStrings(HeadPtr, "DIST"))
               {
               DisplayParam->bnd1 = 1e-8*dblstr(&TailPtr);
               DisplayParam->bnd2 = 1e-8*dblstr(&TailPtr);

               /*   Read in bond colors  */
               DisplayParam->NumBondColor = intstr (&TailPtr);

               /*  Test for proper number of colors  */
               if (DisplayParam->NumBondColor > MAX_COLOR_MAP)
                  {
                  ShutdownGraphics();
                  printf ("ERROR with BOND COLOR.  Too many colors (%i).\n",
                     DisplayParam->NumBondColor);
                  printf ("Maximum allowed is %i.\n", MAX_COLOR_MAP);
                  ShutdownProgram();
                  }

               /*  Read in color list  */
               LOOP (ibond, DisplayParam->NumBondColor)
                  {

                  /*  Read and store color  */
                  ColorStr   = strhed (&TailPtr);
                  Color      = ColorFromStr (ColorStr);
                  AddColorToPalette (Color, &ColorIndex);
                  DisplayParam->BondColorMap[ibond]
                      = ColorIndex;
                  }
               }

            else
               {
               ShutdownGraphics();
               printf ("ERROR with BOND COLOR command.\n");
               printf ("Unknown option.\n");
               ShutdownProgram();
               }

            }

         /*  Bond lengths, default color  */
         else
            {
            DisplayParam->bnd1 = 1e-8*dblstr(&HeadPtr);
            DisplayParam->bnd2 = 1e-8*dblstr(&TailPtr);
            }
         }

      /*  Color atoms  */
      else if (CompareStrings(HeadPtr, "ATOM"))
         {

         HeadPtr = strhed (&TailPtr);

         /*  Turn atom display on  */
         if      (CompareStrings(HeadPtr, "OFF"))
            DisplayParam->AtomFlag = FALSE;

         /*  Turn atom display off  */
         else if (CompareStrings(HeadPtr, "ON"))
            DisplayParam->AtomFlag = TRUE;

         /*  Display atom number rather than image  */
         else if (CompareStrings(HeadPtr, "INDEX"))
            DisplayParam->AtomIndexFlag = TRUE;

         /*  Specify Atom Colors  */
         else if (CompareStrings(HeadPtr, "COLOR"))
            {

            /*  Get color specification  */
            HeadPtr = strhed (&TailPtr);

            /*  Color by type  */
            if      (CompareStrings(HeadPtr, "TYPE"))
               {
               DisplayParam->ColorAtomByTypeFlag = TRUE;
               NumType = intstr (&TailPtr);
               LOOP (itype, NumType)
                  {

                  /*  Atom type  */
                  TempType   = intstr (&TailPtr)-1;

                  /*  Insure that type is between 0 and 255  */
                  TempType  %= MAX_COLOR_MAP;

                  /*  Read and store color  */
                  ColorStr   = strhed (&TailPtr);
                  Color      = ColorFromStr (ColorStr);
                  AddColorToPalette (Color, &ColorIndex);
                  DisplayParam->AtomColorMap[TempType]
                      = ColorIndex;
                  }
               }

            /*  Color by Layer  */
            else if (CompareStrings(HeadPtr, "LAYER"))
               {

               DisplayParam->ColorAtomByLayerFlag = TRUE;

               /*  Read in color list  */
               DisplayParam->NumLayerColor = intstr(&TailPtr);
               LOOP (icolor, DisplayParam->NumLayerColor)
                  {
                  ColorStr   = strhed (&TailPtr);
                  Color      = ColorFromStr (ColorStr);
                  AddColorToPalette (Color, &ColorIndex);
                  DisplayParam->AtomColorMap[icolor] = ColorIndex;
                  }

               }

            /*  Color by displacement  */
            else if (CompareStrings(HeadPtr, "DISP"))
               {
               TempPtr = strhed(&TailPtr);


               /*  Color by displacement angle  */
               if      (CompareStrings(TempPtr,"ANGLE"))
                  {
                  DisplayParam->DispMag = 1e-8*dblstr(&TailPtr);
                  DisplayParam->Phase   = dblstr(&TailPtr);
                  DisplayParam->ColorAtomByDispAngleFlag = TRUE;
                  }

               /*  Color by absolute displacement  */
               else if (CompareStrings(TempPtr,"MAG"))
                  {
                  DisplayParam->DispMag = 1e-8*dblstr(&TailPtr);
                  DisplayParam->ColorAtomByDispMagFlag = TRUE;
                  if (DisplayParam->DispMag== 0)
                     {
                     ShutdownGraphics();
                     printf ("ERROR with DISP MAG command.\n");
                     printf ("  The input Magnitude cannot be zero.\n");
                     ShutdownProgram();
                     }
                  }

               /*  Color by absolute inplane displacement  */
               else if (CompareStrings(TempPtr,"PMAG"))
                  {
                  DisplayParam->DispMag = 1e-8*dblstr(&TailPtr);
                  DisplayParam->ColorAtomByInplaneDispMagFlag = TRUE;
                  if (DisplayParam->DispMag== 0)
                     {
                     ShutdownGraphics();
                     printf ("ERROR with DISP PMAG command.\n");
                     printf ("  The input Magnitude cannot be zero.\n");
                     ShutdownProgram();
                     }
                  }


               /*  Color relative to maximum displacement  */
               else if (CompareStrings(TempPtr,"RMAG"))
                  {
                  DisplayParam->ColorAtomByRelDispMagFlag = TRUE;
                  }

               /*  Color relative to maximum displacement  */
               else if (CompareStrings(TempPtr,"PRMAG"))
                  {
                  DisplayParam->ColorAtomByInplaneRelDispMagFlag = TRUE;
                  }


               /*  Color by dot product with vector  */
               else if (CompareStrings(TempPtr,"VECT"))
                  {
                  DisplayParam->vectmin = 1e-8*dblstr(&TailPtr);
                  DisplayParam->vectmax = 1e-8*dblstr(&TailPtr);

                  DisplayParam->vector[X] = dblstr(&TailPtr);
                  DisplayParam->vector[Y] = dblstr(&TailPtr);
                  DisplayParam->vector[Z] = dblstr(&TailPtr);
                  DisplayParam->ColorAtomByDispVectFlag = TRUE;
                  }

               /*  Color relative to maximum dot product with vector  */
               else if (CompareStrings(TempPtr,"RVECT"))
                  {
                  DisplayParam->vector[X] = dblstr(&TailPtr);
                  DisplayParam->vector[Y] = dblstr(&TailPtr);
                  DisplayParam->vector[Z] = dblstr(&TailPtr);
                  DisplayParam->ColorAtomByRelDispVectFlag = TRUE;
                  }

               /*  Unknown ATOM COLOR DISP option  */
               else
                  {
                  printf
                     ("ERROR:  Unknown option for ATOM COLOR DISP command.\n");
                  WarningCount++;
                  }

               /*  Read in color list  */
               DisplayParam->NumDispColor = intstr(&TailPtr);
               LOOP (icolor, DisplayParam->NumDispColor)
                  {
                  ColorStr   = strhed (&TailPtr);
                  Color      = ColorFromStr (ColorStr);
                  AddColorToPalette (Color, &ColorIndex);
                  DisplayParam->AtomColorMap[icolor] = ColorIndex;
                  }
               }

            /*  Unknown ATOM COLOR option  */
            else
               {
               printf ("ERROR:  Unknown option for ATOM COLOR command.\n");
               WarningCount++;
               }
            }

         /*  Specify Atom Radii */
         else if (CompareStrings(HeadPtr, "RADIUS"))
            {

            /*  Get color specification  */
            HeadPtr = strhed (&TailPtr);

            /*  Color by type  */
            if      (CompareStrings(HeadPtr, "TYPE"))
               {
               DisplayParam->RadiusAtomByTypeFlag = TRUE;
               NumType = intstr (&TailPtr);
               LOOP (itype, NumType)
                  {

                  /*  Atom type  */
                  TempType   = intstr (&TailPtr)-1;

                  /*  Insure that type is between 0 and 255  */
						TempType  %= MAX_RADIUS_MAP;

                  /*  Read and store Radius */
                  DisplayParam->AtomRadiusMap[TempType]
                      = dblstr(&TailPtr);
                  }
               }
				}


         else
            {
            printf ("ERROR:  Unknown option for ATOM command.\n");
            WarningCount++;
            }
         }
         /*  else if (CompareStrings(HeadPtr, "ATOM"))  */


      /*  Specify displacement type  */
      else if (CompareStrings(HeadPtr, "DISP"))
         {
         DisplayParam->DispFlag = TRUE;

         HeadPtr = strhed (&TailPtr);

         /*  Do displacement relative to first step in file  */
         if (CompareStrings(HeadPtr, "REF"))
            {
            DisplayParam->DispRefFlag = TRUE;
            HeadPtr = strhed (&TailPtr);
            ALLOCATE(DisplayParam->DispRefFileName, char, strlen(HeadPtr)+1);
            strcpy  (DisplayParam->DispRefFileName, HeadPtr);
            HeadPtr = strhed (&TailPtr);
            }

         /*  Check for slip option  */
         if (CompareStrings(HeadPtr, "SLIP"))
            {
            DisplayParam->TailSlipFlag = TRUE;
            DisplayParam->TailSlipRad  = 1e-8*dblstr (&TailPtr);
            HeadPtr = strhed (&TailPtr);
            }


         }


      /*  Fixed frame  */
      else if (CompareStrings(HeadPtr, "FRAME"))
         {
         DisplayParam->FixedFrameFlag = TRUE;
         DisplayParam->Frame[0][X] = 1e-8*dblstr(&TailPtr);
         DisplayParam->Frame[0][Y] = 1e-8*dblstr(&TailPtr);
         DisplayParam->Frame[0][Z] = 1e-8*dblstr(&TailPtr);
         DisplayParam->Frame[1][X] = 1e-8*dblstr(&TailPtr);
         DisplayParam->Frame[1][Y] = 1e-8*dblstr(&TailPtr);
         DisplayParam->Frame[1][Z] = 1e-8*dblstr(&TailPtr);
         SWAP(DisplayParam->Frame[0][X], DisplayParam->Frame[1][X], TempDouble)
         SWAP(DisplayParam->Frame[0][Y], DisplayParam->Frame[1][Y], TempDouble)
         SWAP(DisplayParam->Frame[0][Z], DisplayParam->Frame[1][Z], TempDouble)

         /*  Sanity Check  */
         if
         (
         (DisplayParam->Frame[0][X]==DisplayParam->Frame[1][X]) ||
         (DisplayParam->Frame[0][Y]==DisplayParam->Frame[1][Y]) ||
         (DisplayParam->Frame[0][Z]==DisplayParam->Frame[1][Z])
         )
            {
            ShutdownGraphics();
            printf ("ERROR:  FRAME has zero volume.\n");
            ShutdownProgram();
            }
         }


      /*  Draw tails  */
      else if (CompareStrings(HeadPtr, "TAIL"))
         {

         /*  Set flag for drawing tails  */
         DisplayParam->TailFlag = TRUE;
         HeadPtr = strhed (&TailPtr);

         /*  Read scale for plotting tails  */
			if (IsBlank(HeadPtr))
				DisplayParam->TailScale = 1.0;
			else
	         DisplayParam->TailScale = dblstr (&HeadPtr);
         }


      /*  read file title  */
      else if (CompareStrings(HeadPtr, "TITLE"))
         {

         /*  Get number of lines in title  */
         DisplayParam->NumTitle = intstr (&TailPtr);

         /*  Test for NumTitle>0  */
         if (DisplayParam->NumTitle <= 0)
            {
            ShutdownGraphics();
            printf ("ERROR:  TITLE command.\n");
            printf ("Number of lines in title must be greater than 0.\n");
            ShutdownProgram();
            }

         /*  Allocate space for pointers to title lines  */
         ALLOCATE (DisplayParam->Title, char *, DisplayParam->NumTitle)

         /*  Read title lines  */
         ititle = 0;
         while
         (
         ititle<DisplayParam->NumTitle &&
         NULL!=fgets (CmdFileBuffer, STR_LEN, InputFile)
         )
            {

				/*  Remove trailing \n  */
				TruncateString (CmdFileBuffer);

            /*  Find length of string (without trailing \n)  */
            StringLength = strlen (CmdFileBuffer)-1;

            /*  Allocate space for string   */
            ALLOCATE
               (DisplayParam->Title[ititle], char, StringLength+1)

            /*  Copy string  */
            strcpy (DisplayParam->Title[ititle], CmdFileBuffer);

            ititle++;

            }
         }

      /* get the resolution of the screen */
      else if (CompareStrings(HeadPtr,"RESOLUTION"))
         {
         DisplayParam->hsmax = intstr(&TailPtr)-1;
         DisplayParam->vsmax = intstr(&TailPtr)-1;
         }



      /* get the resolution of the screen */
      else if (CompareStrings(HeadPtr,"VIEW"))
         {
         ViewAxis[X] = dblstr(&TailPtr);
         ViewAxis[Y] = dblstr(&TailPtr);
         ViewAxis[Z] = dblstr(&TailPtr);
         VertAxis[X] = dblstr(&TailPtr);
         VertAxis[Y] = dblstr(&TailPtr);
         VertAxis[Z] = dblstr(&TailPtr);

         /*  Calculate rotation matrix, and input validity  */
         InitializeView(DisplayParam, ViewAxis, VertAxis);

         }



      /* unknown input */
      else
         {
         printf("Error in File\nUnknown Command Line: '%s'",
            CmdFileBuffer);
         WarningCount++;
         }

      } /*while not eof*/


   fclose(InputFile);

#ifdef GRAPHICS
   /*  If warnings encountered, allow user to view them  */
   if (WarningCount!=0)
      {
      printf ("\n");
      printf ("Hit enter to continue.\n");
      getch  ();
      }
#endif

   }


/*  Read command line options  */
void ReadCommandLine
(
int    argc,
char   *argv[],
dsparm *DisplayParam,
WORD2   *NumTokensRead
)
   {
   int iopt;
   char c;

    /*   Parse while parameter has - and is not a number  */
   iopt = 1;
   while (     iopt<argc
          && argv[iopt][0]=='-'
          && ( (c=argv[iopt][1])<'0' || c>'9')
         )

      {

      switch (c)
         {

         /*  Read type file  */
         case 't':
            DisplayParam->TypeListFlag = TRUE;
            DisplayParam->TypeListFileName = argv[iopt+1];
            iopt += 1;
            break;

         /*  Draw bonds  */
         case 'b':
            DisplayParam->BondFlag = TRUE;
            DisplayParam->bnd1 = atof (argv[iopt+1]);
            DisplayParam->bnd2 = atof (argv[iopt+2]);
            iopt += 2;
            break;

         /*  Draw displacement relative to constant step  */
         case 'd':
            DisplayParam->TailFlag        = TRUE;
            DisplayParam->DispFlag        = TRUE;
            DisplayParam->DispRefFlag     = TRUE;
            DisplayParam->DispRefFileName = argv[iopt+1];
            DisplayParam->TailScale = atof (argv[iopt+2]);
            iopt   += 2;
            break;

         /*  Draw displacement relative to constant step  */
         case 'm':
            DisplayParam->TailFlag    = TRUE;
            DisplayParam->DispFlag    = TRUE;
            DisplayParam->DispRefFlag = FALSE;
            DisplayParam->TailScale = atof (argv[iopt+1]);
            iopt   += 1;
            break;

         /*  Scale displacements  */
         case 's':
            DisplayParam->TailSlipFlag = TRUE;
            DisplayParam->TailSlipRad  = atof   (argv[iopt+1]);
            iopt   += 1;
            break;

         /*  Wrap bonds around repeating boundary conditions  */
         case 'w':
            DisplayParam->BondWrapFlag = TRUE;
            break;

         /*  Draw particle numbers instead of symbols */
         case 'x':
            DisplayParam->AtomIndexFlag = TRUE;
            break;

         /*  Draw NO particle symbols or numbers  */
         case 'n':
            DisplayParam->AtomFlag      = FALSE;
            break;

         /*  Reference file name  */
         case 'r':
            DisplayParam->LayerRefFlag = TRUE;
            DisplayParam->LayerRefFileName = argv[iopt+1];
            iopt++;
            break;

         default :
            fprintf (stderr,
               "Command line switch %s unknown.\n", argv[iopt]);
            break;
         }
      iopt++;
      }

   /*  Returns number of tokens read  */
   *NumTokensRead = iopt-1;
   }


void ParseOrientStr (char *InputStr, int *hor, int *ver, int *dep)
   {
   *ver = ToUpper(InputStr[0])-'X';
   *hor = ToUpper(InputStr[2])-'X';
   if (*ver==*hor || *hor>2 ||  *ver>2 || *hor<0 || *ver<0)
      {
      ShutdownGraphics();
      printf ("ERROR:  Invalid orientation \"%s\"\n", InputStr);
      ShutdownProgram();
      }
   *dep = 3- (*ver) - (*hor);
   }



void InitializeDisplayParam (dsparm *DisplayParam)
   {
   int idir;
   int jdir;

   /*  SET draw step parameters  */
   DisplayParam->aspect    = DEF_ASPECT_RATIO;
   DisplayParam->LabelFlag = TRUE;
   DisplayParam->hsmin =   0;
   DisplayParam->vsmin =   0;

   /*  NOTE:  DisplayParam->hsmax and vsmax are set below  */
   DisplayParam->tmarg =  10;
   DisplayParam->bmarg =  10;
   DisplayParam->lmarg =   5;
   DisplayParam->rmarg =   5;
   DisplayParam->timarg = 10;
   DisplayParam->bimarg = 10;
   DisplayParam->limarg = 10;
   DisplayParam->rimarg = 10;

   /*
   Initialize resolution as 0,0.
   If not changed before graphics mode set,
   program will change to default.
   */
   DisplayParam->hsmax  =  0;
   DisplayParam->vsmax  =  0;

   /*  Flag for drawing atoms  */
   DisplayParam->AtomFlag                    = TRUE;

   DisplayParam->AtomIndexFlag               = FALSE;
   DisplayParam->BondFlag                    = FALSE;
   DisplayParam->BondWrapFlag                = FALSE;
   DisplayParam->ColorAtomByDispAngleFlag    = FALSE;
   DisplayParam->ColorAtomByDispMagFlag      = FALSE;
   DisplayParam->ColorAtomByInplaneDispMagFlag      = FALSE;
   DisplayParam->ColorAtomByDispVectFlag     = FALSE;
   DisplayParam->ColorAtomByRelDispMagFlag   = FALSE;
   DisplayParam->ColorAtomByInplaneRelDispMagFlag   = FALSE;
   DisplayParam->ColorAtomByRelDispVectFlag  = FALSE;
   DisplayParam->ColorAtomByTypeFlag         = FALSE;
   DisplayParam->ColorBondByDistFlag         = FALSE;
   DisplayParam->DispFlag                    = FALSE;
   DisplayParam->DispRefFlag                 = FALSE;
   DisplayParam->LayerRefFlag                = FALSE;
   DisplayParam->TailFlag                    = FALSE;
   DisplayParam->TailSlipFlag                = FALSE;
   DisplayParam->TypeListFlag                = FALSE;

   DisplayParam->Title                       = NULL;
   DisplayParam->NumTitle                    = 0;

   DisplayParam->FixedFrameFlag              = FALSE;
   DisplayParam->ViewFlag                    = FALSE;

   /*  Initialize rotation matrix to identity  */
   LOOP(idir,NDIR)
   LOOP(jdir,NDIR)
      {
      DisplayParam->Rotate[idir][jdir] = 0.0;
      }
   LOOP(idir,NDIR)
      {
      DisplayParam->Rotate[idir][idir] = 1.0;
      }
   }



/*  Normalize 3 vector  */
void Normalize (float Vector[NDIR])
   {
   float MagSqrd;

   MagSqrd =
      Vector[X]*Vector[X] +
      Vector[Y]*Vector[Y] +
      Vector[Z]*Vector[Z];

   if (MagSqrd==0.0)
      return;

   Vector[X] /= MagSqrd;
   Vector[Y] /= MagSqrd;
   Vector[Z] /= MagSqrd;
   }


/*  Compare two strings ignoring case.  Return true of identical  */
BOOLEAN CompareStrings (char *FirstStr, char *SecondStr)
   {
   BOOLEAN CharsAreEqual;
   char FirstChar;
   char SecondChar;

   FirstChar  = ToUpper (*FirstStr );
   SecondChar = ToUpper (*SecondStr);
   CharsAreEqual = (FirstChar==SecondChar);

   while (CharsAreEqual && FirstStr[0]!='\0')
      {
      FirstStr++;
      SecondStr++;
      FirstChar  = ToUpper (*FirstStr );
      SecondChar = ToUpper (*SecondStr);
      CharsAreEqual = (FirstChar==SecondChar);
      }

   /*  Strings equal if CharsAreEqual==TRUE  */
   return CharsAreEqual;

   }



void InitializeView
(
dsparm *DisplayParam,
double View[NDIR],
double Vert[NDIR]
)
   {
   double Cross[NDIR];
   double Mag;

   /*  Calculate direction orthgonal to both View Axis and Vert Axis  */
   CROSS(Cross, View, Vert)

   /*  Calculate Magnitude  */
   Mag = Cross[X]*Cross[X] + Cross[Y]*Cross[Y] + Cross[Z]*Cross[Z];

   /*  Two axis are parallel  */
   if (Mag==0)
      {
      ShutdownGraphics();
      printf ("ERROR:  The two axis for the VIEW command are parallel.\n");
      ShutdownProgram();
      }

   /*  Orthogonalize Vert[] to other two vectors  */
   CROSS(Vert, Cross, View)

   /*  Normalize all vectors  */
   NORMALIZE(View)
   NORMALIZE(Vert)
   NORMALIZE(Cross)

   /*  Calculate Rotational View Matrix  */
   CROSS (DisplayParam->Rotate[X], Vert,  View)
   CROSS (DisplayParam->Rotate[Y], View,  Cross)
   CROSS (DisplayParam->Rotate[Z], Cross, Vert)

   /*  Set flag  */
   DisplayParam->ViewFlag = TRUE;
   }


void TruncateString (char *InputStr)
	{
	int LastPos;

	/*  Skip if null string  */
	if (InputStr[0]=='\0')
		return;

	LastPos = strlen(InputStr) - 1;

	if (InputStr[LastPos]=='\n')
		InputStr[LastPos] = '\0';
	}
